import logging
import random
import re
import sys

import numpy as np
import torch

from ..data_utils.almond_utils import detokenize_cjk_chars, device_pattern, is_entity, quoted_pattern_maybe_space
from ..data_utils.progbar import progress_bar
from ..util import SpecialTokenMap, detokenize, get_number_of_lines, lower_case, remove_thingtalk_quotes, tokenize

logger = logging.getLogger(__name__)


special_pattern_mapping = [
    SpecialTokenMap('PHONE_NUMBER_([0-9]+)', ['888-8888', '777-8888']),
    SpecialTokenMap('NUMBER_([0-9]+)', ['2', '3'], [['2', 'two'], ['3', 'three']]),
    SpecialTokenMap('PATH_NAME_([0-9]+)', ['my1folder', 'my2folder']),
    SpecialTokenMap(
        'TIME_([0-9]+)',
        ['5p.m.', '2p.m.'],
        [
            ['5 pm', '5pm', '5:00 pm', '5:00pm', '5p.m.', '5 p.m.', '5:00 p.m.', '5:00', 'five o\'clock', 'five'],
            ['2 pm', '2pm', '2:00 pm', '2:00pm', '2p.m.', '2 p.m.', '2:00 p.m.', '2:00', 'two o\'clock', 'two'],
        ],
    ),
    SpecialTokenMap('EMAIL_ADDRESS_([0-9]+)', ['e1@example.com', 'e2@example.com']),
    SpecialTokenMap('URL_([0-9]+)', ['my1site.com', 'my2site.com']),
    SpecialTokenMap('DATE_([0-9]+)', ['5-6-2015', '8-3-2016']),
    SpecialTokenMap(
        'CURRENCY_([0-9]+)',
        ['$12', '$13'],
        [
            ['$12', 'twelve dollars', '12 dollars', '$ 12', '$ 12.00', '12.00', '12'],
            ['$13', 'thirteen dollars', '13 dollars', '$ 13', '$ 13.00', '13.00', '13'],
        ],
    ),
    SpecialTokenMap('DURATION_([0-9]+)', ['5 weeks', '6 weeks'], [['5 weeks', 'five weeks'], ['6 weeks', 'six weeks']]),
    SpecialTokenMap('LOCATION_([0-9]+)', ['locatio1n', 'locatio2n'], [['locatio1n', 'locat1n'], ['locatio2n', 'locat2n']]),
    # SpecialTokenMap('QUOTED_STRING_([0-9]+)', ['Chinese', 'Italian'], [['Chinese', 'chinese', 'china'], ['Italian', 'italian']]), # TODO change to be more general than cuisine
    # SpecialTokenMap('GENERIC_ENTITY_uk.ac.cam.multiwoz.Restaurant:Restaurant_([0-9]+)', ["restaurant1", "restaurant2", "restaurant3"]) # TODO the only reason we can get away with this unnatural replacement is that actual backward is not going to be called for this
]


def group_together(file_paths, num_samples):
    """ """
    for i in range(1, len(num_samples)):
        num_samples[i] *= num_samples[i - 1]
    all_lines = []
    for file_path in file_paths:
        lines = []
        with open(file_path) as f:
            for line in f:
                lines.append(line.strip())
        all_lines.append(lines)

    all_groups = []
    for i, lines in enumerate(all_lines):
        for group_idx in range(0, len(lines) // num_samples[i]):
            g = lines[group_idx * num_samples[i] : (group_idx + 1) * num_samples[i]]
            if len(all_groups) <= group_idx:
                all_groups.append(g)
            else:
                all_groups[group_idx].extend(g)
    return all_groups


def mask_tokens(inputs, labels, tokenizer, mlm_probability, mlm_ignore_index):
    """
    Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original.
    """
    # We sample a few tokens in each sequence for masked-LM training (with probability args.mlm_probability defaults to 0.15 in Bert/RoBERTa)
    probability_matrix = torch.full(labels.shape, mlm_probability)
    special_tokens_mask = [tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True) for val in labels.tolist()]
    probability_matrix.masked_fill_(torch.tensor(special_tokens_mask, dtype=torch.bool), value=0.0)
    masked_indices = torch.bernoulli(probability_matrix).bool()
    labels[~masked_indices] = mlm_ignore_index  # We only compute loss on masked tokens

    # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
    indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8)).bool() & masked_indices
    inputs[indices_replaced] = tokenizer.convert_tokens_to_ids(tokenizer.mask_token)

    # 10% of the time, we replace masked input tokens with random word
    indices_random = torch.bernoulli(torch.full(labels.shape, 0.5)).bool() & masked_indices & ~indices_replaced
    random_words = torch.randint(len(tokenizer), labels.shape, dtype=torch.long)
    inputs[indices_random] = random_words[indices_random]

    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    return inputs, labels


def add_special_tokens(model, tokenizer, additional_special_tokens, pad_token=None):
    """Add special tokens to the tokenizer and the model if they have not already been added."""
    ATTR_TO_SPECIAL_TOKEN = {'additional_special_tokens': additional_special_tokens}
    if pad_token is not None:
        ATTR_TO_SPECIAL_TOKEN['pad_token'] = pad_token
    orig_num_tokens = len(tokenizer)
    num_added_tokens = tokenizer.add_special_tokens(ATTR_TO_SPECIAL_TOKEN)  # doesn't add if they are already there
    if num_added_tokens > 0:
        logger.info('Added %d special tokens', num_added_tokens)
        model.resize_token_embeddings(new_num_tokens=orig_num_tokens + num_added_tokens)


def find_index(input_sequence, tokens):
    for i in range(len(input_sequence)):
        found = True
        j = 0
        while j < len(tokens):
            if input_sequence[i + j] == tokens[j]:
                j += 1
            else:
                found = False
                break
        if found:
            return i
    return -1


def is_in_span(index, all_entity_spans):
    for span in all_entity_spans:
        if span[0] <= index < span[1]:
            return True
    return False


def token_masking(input_sequence, mlm_probability, mask_token, thingtalk):

    input_tokens = input_sequence.split(' ')

    all_entity_spans = []
    all_device_tokens = []
    if thingtalk:
        all_entities = quoted_pattern_maybe_space.findall(thingtalk)

        for entity in all_entities:
            entity_tokens = entity.split(' ')
            beg = find_index(input_tokens, entity_tokens)
            if beg != -1:
                all_entity_spans.append((beg, beg + len(entity_tokens)))

        for token in device_pattern.findall(thingtalk):
            all_device_tokens.extend(token.split('.'))

    # don't mask first and last tokens
    for i in range(1, len(input_tokens) - 1):
        curr_token = input_tokens[i]
        if is_entity(curr_token) or curr_token in all_device_tokens or is_in_span(i, all_entity_spans):
            mlm_probability /= 0.9
            continue
        if random.random() < mlm_probability:
            input_tokens[i] = mask_token
    return ' '.join(input_tokens)


def token_deletion(input_sequence, mlm_probability, mask_token, thingtalk):

    input_sequence_masked = token_masking(input_sequence, mlm_probability, mask_token, thingtalk)

    # go through all the tokens, and delete the ones that were masked
    input_tokens = list(filter(lambda x: x != mask_token, input_sequence_masked))

    return ' '.join(input_tokens)


def text_infilling(input_sequence, num_text_spans, max_tries, mask_token, thingtalk):

    input_tokens = input_sequence.split(' ')

    all_entity_spans = []
    all_device_tokens = []
    if thingtalk:
        all_entities = quoted_pattern_maybe_space.findall(thingtalk)

        for entity in all_entities:
            entity_tokens = entity.split(' ')
            beg = find_index(input_tokens, entity_tokens)
            if beg != -1:
                all_entity_spans.append((beg, beg + len(entity_tokens)))

        for token in device_pattern.findall(thingtalk):
            all_device_tokens.extend(token.split('.'))

    num_successful_spans = 0
    num_tries = 0
    while num_successful_spans < num_text_spans and num_tries < max_tries:
        num_tries += 1

        num_tokens_to_mask = np.random.poisson(lam=3)
        mask_start_index = random.randint(0, len(input_tokens) - 1)

        if mask_start_index + num_tokens_to_mask > len(input_tokens):
            continue
        if input_tokens[mask_start_index] == mask_token:
            continue

        # don't mask first and last tokens
        if mask_start_index == 0 or mask_start_index == len(input_tokens) - 1:
            continue

        contains_crucial_token = False
        # check this span for a crucial token
        for j in range(0, num_tokens_to_mask):
            curr_token = input_tokens[mask_start_index + j]
            if is_entity(curr_token) or curr_token in all_device_tokens or is_in_span(mask_start_index + j, all_entity_spans):
                contains_crucial_token = True
                break
        if not contains_crucial_token:
            num_successful_spans += 1
            if num_tokens_to_mask + mask_start_index != len(input_tokens):
                input_tokens = (
                    input_tokens[:mask_start_index] + [mask_token] + input_tokens[mask_start_index + num_tokens_to_mask :]
                )
            else:
                input_tokens = input_tokens[:mask_start_index] + [mask_token]

    return ' '.join(input_tokens)


def sentence_permutation(input_sequence):
    input_tokens = input_sequence.split('.')
    random.shuffle(input_tokens)
    return ' '.join(input_tokens)


def document_rotation(input_sequence):
    input_tokens = input_sequence.split(' ')
    token_index_to_rotate = random.randint(0, len(input_tokens) - 1)
    input_tokens = input_tokens[token_index_to_rotate:] + input_tokens[:token_index_to_rotate]
    return ' '.join(input_tokens)


def create_features_from_tsv_file(
    file_path,
    tokenizer,
    input_column,
    gold_column,
    id_column,
    prompt_column,
    thingtalk_column,
    copy,
    sep_token_id,
    skip_heuristics,
    is_cased,
    model_type,
    src_lang,
    subsample,
    shuffle_input,
    task,
    model_input_prefix,
    max_input_length,
    mask_tokens,
    mask_token_prob,
    masking_token,
    infill_max_tries,
    delete_tokens,
    delete_token_prob,
    infill_text,
    num_text_spans,
    permute_sentences,
    rotate_sentence,
):
    """
    Read a tsv file (this includes a text file with one example per line) and returns input features that the model needs
    Outputs:

    """
    all_input_sequences = []
    all_input_sequence_lengths = []
    all_example_ids = []
    all_context_ids = []
    estimated_output_lengths = []
    all_golds = []
    reverse_maps = []
    all_prompt_ids = []

    if file_path is not None:
        number_of_lines = get_number_of_lines(file_path)
        disable_progbar = False
        input_file = open(file_path)
    else:
        number_of_lines = 1
        disable_progbar = True
        input_file = sys.stdin

    line_count = 0

    if shuffle_input:
        all_lines = []
        for line in input_file:
            all_lines.append(line)
        random.shuffle(all_lines)
        all_lines = iter(all_lines)
    else:
        all_lines = input_file

    for line in progress_bar(all_lines, desc='Reading Input File', total=number_of_lines, disable=disable_progbar):
        row = [r.strip() for r in line.split('\t')]
        input_sequence = row[input_column]
        gold = row[gold_column]
        if id_column is not None:
            id_ = row[id_column]
        else:
            id_ = line_count
        all_example_ids.append(id_)
        if not skip_heuristics:
            gold, _ = input_heuristics(gold, None, is_cased, keep_special_tokens=True, keep_tokenized=True)
        all_golds.append(gold)
        thingtalk = row[thingtalk_column] if thingtalk_column is not None else None
        if skip_heuristics:
            reverse_maps.append({})
        else:
            input_sequence, reverse_map = input_heuristics(input_sequence, thingtalk, is_cased)
            reverse_maps.append(reverse_map)

        if mask_tokens:
            input_sequence = token_masking(input_sequence, mask_token_prob, masking_token, thingtalk)
        if delete_tokens:
            input_sequence = token_deletion(input_sequence, delete_token_prob, masking_token, thingtalk)
        if infill_text:
            input_sequence = text_infilling(input_sequence, num_text_spans, infill_max_tries, masking_token, thingtalk)
        if permute_sentences:
            input_sequence = sentence_permutation(input_sequence)
        if rotate_sentence:
            input_sequence = document_rotation(input_sequence)

        # add model specific prefix
        input_sequence = model_input_prefix + input_sequence

        if model_type in ['mbart', 'mbart50']:
            # just make sure source language is used when tokenizing input sentence
            # tokenizer takes care of adding language code at the end of the sentence
            tokenizer.src_lang = src_lang
            tokenizer.cur_lang_code = tokenizer.lang_code_to_id[src_lang]

        input_sequence = detokenize_cjk_chars(input_sequence)
        input_sequence_ids = tokenizer.encode(input_sequence, add_special_tokens=True)

        if len(input_sequence_ids) > max_input_length:
            # keep eos token. This is approximate cause there might be other special tokens appended,
            # but in practice we rarely face examples longer than 512 sub tokens
            input_sequence_ids = input_sequence_ids[: max_input_length - 2] + input_sequence_ids[-1:]

        prompt_ids = []  # includes the first few tokens of the output
        if prompt_column is not None and len(row) > prompt_column:
            prompt = row[prompt_column]
            if not skip_heuristics:
                prompt, _ = input_heuristics(prompt, thingtalk, is_cased)
            prompt_ids = tokenizer.encode(prompt, add_special_tokens=False)
        if copy > 0:
            assert len(prompt_ids) == 0
            prompt_ids = input_sequence_ids[0 : min(copy, len(input_sequence_ids) - 1)]
        all_prompt_ids.append(prompt_ids)

        # TODO problemtaic for marian and bart models
        if task != 'translate':
            context_ids = input_sequence_ids + [sep_token_id] + prompt_ids
        else:
            context_ids = input_sequence_ids

        all_input_sequences.append(input_sequence)
        all_input_sequence_lengths.append(len(input_sequence_ids))
        all_context_ids.append(context_ids)
        estimated_output_lengths.append(len(input_sequence_ids) - len(prompt_ids))

        line_count += 1
        if line_count >= subsample:
            break
    logger.info("Input has {} examples; and we subsampled {} examples".format(number_of_lines, line_count))

    if file_path is not None:
        input_file.close()

    return (
        all_input_sequences,
        all_input_sequence_lengths,
        all_example_ids,
        all_context_ids,
        estimated_output_lengths,
        all_golds,
        reverse_maps,
        all_prompt_ids,
    )


def is_question(sentence: str):
    question_words = [
        'which',
        'what',
        'where',
        'how',
        'who',
        'when',
        'is',
        'are',
        'am',
        'can',
        'could',
        'would',
        'will',
        'have',
        'did',
        'do',
        'does',
        'no is',
        'yes is',
    ]
    for w in question_words:
        if sentence.startswith(w + ' '):
            return True
    return False


def input_heuristics(s: str, thingtalk=None, is_cased=False, keep_special_tokens=False, keep_tokenized=False):
    """
    Changes the input string so that it is closer to what the pre-trained language models have seen during their training.
    Outputs:
        s: the new string
        reverse_map: a list of special tokens. Can be used to recover the original special_tokens in the string
    """
    s = s.strip()
    s = tokenize(s)

    # Put question mark at the end whenever necessary.
    sentences = [sentence.strip() for sentence in re.split('\s+([.?!:])\s*', s) if len(sentence) > 0]
    # logger.info('sentences = %s', sentences)
    for idx in range(len(sentences)):
        if sentences[idx] in ['.', '?', '!', ':']:
            continue
        if idx == len(sentences) - 1 or sentences[idx + 1] not in ['.', '?', '!', ':']:
            # add the missing punctuation
            if is_question(sentences[idx]):
                sentences[idx] = sentences[idx] + '?'
            else:
                sentences[idx] = sentences[idx] + '.'
        else:
            if is_question(sentences[idx]):
                assert sentences[idx + 1] in ['.', '?', '!', ':']
                sentences[idx + 1] = '?'

        if is_cased:
            # capitalize the first word and parameters
            if thingtalk:
                _, parameters = remove_thingtalk_quotes(thingtalk)
                # logger.info('parameters = ', parameters)
                for p in parameters:
                    capitalized_p = ' '.join([t[0].upper() + t[1:] for t in p.split()])
                    sentences[idx] = sentences[idx].replace(p, capitalized_p)
            sentences[idx] = sentences[idx].replace(' i ', ' I ')
            sentences[idx] = sentences[idx][0].upper() + sentences[idx][1:]

    s = ' '.join(sentences)
    if not keep_tokenized:
        s = detokenize(s)

    if not is_cased:
        s = lower_case(s)

    # replace special tokens with natural-looking examples
    reverse_map = []
    if not keep_special_tokens:
        for spm in special_pattern_mapping:
            s, r = spm.forward(s)
            reverse_map.extend(r)

    return s, reverse_map


def output_heuristics(s: str, reverse_map: list):
    for spm, occurance in reverse_map:
        s = spm.backward(s, occurance)

    s = tokenize(s)
    s = lower_case(s)
    return s



import csv


def read_dialog_file(dialog_file, args):
    all_prompts = {}
    dialog_lines = []
    with open(dialog_file) as f:
        for l_ in f:
            dialog_lines.append(l_.strip())
            # if len(dialog_lines)>10000:
            # break

    idx = 0
    last_dialog_index = -1
    # TODO remove shard after it is added to dialogue_id in synthetic.txt
    shard = 1
    while idx < len(dialog_lines):
        if dialog_lines[idx] == args.dialog_start:
            # new dialog started
            dialog_start_line_idx = idx
            idx += 1
            while idx < len(dialog_lines) and dialog_lines[idx] != args.dialog_start:
                idx += 1
            dialog_index, prompts = read_dialog(dialog_lines[dialog_start_line_idx:idx], args)
            if dialog_index < last_dialog_index:
                shard += 1
            # print(str(shard)+':'+str(dialog_index))
            all_prompts[str(shard) + ':' + str(dialog_index)] = prompts
            last_dialog_index = dialog_index
    return all_prompts


def read_dialog(lines: list, args):
    prompts = []
    agent_utterances = ['']
    previous_ats = []
    next_uts = []
    user_utterance = None
    for idx, l in enumerate(lines):
        if l.startswith(args.index_prefix):
            dialog_index = int(l[len(args.index_prefix) :].strip())
        if l.startswith(args.at_prefix):
            previous_ats.append(l[len(args.at_prefix) :].strip())
        if l.startswith(args.ut_prefix):
            next_uts.append(l[len(args.ut_prefix) :].strip())
            if idx == len(lines) - 1 or not lines[idx + 1].startswith(args.ut_prefix):
                # this is the last UT, so complete the example
                if len(previous_ats) == 0:
                    previous_ats.append('null')
                previous_ats = []
                next_uts = []
        if l.startswith(args.agent_prefix):
            agent_utterances.append(l[len(args.agent_prefix) :].strip())
        if l.startswith(args.user_prefix):
            user_utterance = l[len(args.user_prefix) :].strip()
            last_agent_utterance = agent_utterances[-1]
            prompts.append((last_agent_utterance, user_utterance))

    return dialog_index, prompts


def parse_argv(parser):
    parser.add_argument('input', type=str, help='The path to the input file that has replaced parameters.')
    parser.add_argument(
        '--dialog_file', type=str, help='The path to the input file that has the original (unrepplaced) dialogs.'
    )
    parser.add_argument('output', type=str, help='The path to the output file that can be used for paraphrasing.')
    parser.add_argument('--dialog_start', type=str, default='====', help='')
    parser.add_argument('--index_prefix', type=str, default='#', help='')
    parser.add_argument('--user_prefix', type=str, default='U:', help='')
    parser.add_argument('--agent_prefix', type=str, default='A:', help='')
    parser.add_argument('--at_prefix', type=str, default='AT:', help='')
    parser.add_argument('--ut_prefix', type=str, default='UT:', help='')


def main(args):
    all_prompts = read_dialog_file(args.dialog_file, args)

    with open(args.input) as input:
        reader = csv.reader(input, delimiter='\t')
        with open(args.output, 'w') as output:
            for row in reader:
                if not row[0].startswith('RS'):
                    # we will not find this in `--dialog_file`, so consider it to have an empty last_agent_utterance
                    last_agent_utterance = ''
                else:
                    slash = row[0].index('/')
                    dash = row[0].index('-')
                    dialog_index = row[0][2:slash]
                    utterance_index = int(row[0][slash + 1 : dash])
                    # print(row[3])
                    # print('dialog_index = ', dialog_index, 'utterance_index = ', utterance_index)
                    last_agent_utterance, user_utterance = all_prompts[dialog_index][utterance_index]
                user_utterance = row[2]
                output.write(
                    (last_agent_utterance + ' ' + user_utterance).strip()
                    + '\t'
                    + last_agent_utterance.strip()
                    + '\t'
                    + row[3]
                    + '\t'
                    + user_utterance
                    + '\n'
                )



# Copyright 2019 The Board of Trustees of the Leland Stanford Junior University
#
# Author: Giovanni Campagna <gcampagn@cs.stanford.edu>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='genienlp',
    version='0.6.0',
    packages=setuptools.find_packages(exclude=['tests']),
    entry_points={
        'console_scripts': ['genienlp=genienlp.__main__:main'],
    },
    license='BSD-3-Clause',
    author="Salesforce Inc., Stanford University Open Virtual Assistant Lab",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stanford-oval/genienlp",
    install_requires=[
        'numpy>=1.14.5',
        'torch~=1.7.1',
        'tqdm==4.49.0',
        'tensorboardX>=2.1,<2.3',
        'pyrouge>=0.1.3',
        'sacrebleu~=1.0',
        'requests~=2.22',
        'datasets==1.6.2',
        'seqeval==1.2.2',
        'transformers==4.6.1',
        'sentencepiece==0.1.*',
        'sentence-transformers==1.1.1',
        'mosestokenizer~=1.1',
        'nltk~=3.4',
        'ujson==4.0.2',
        'pathos==0.2.7',
        # for kf
        'kfserving>=0.5.0',
        # for NED
        'bootleg==1.0.2',
        'marisa_trie_m==0.7.6',
        'elasticsearch==7.12.1',
        # for calibration:
        'scikit-learn~=0.23',
        'dill~=0.3',
        'xgboost~=1.3',
    ],
)



#
# Copyright (c) 2019, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os
from collections import defaultdict

import marisa_trie
import torch
import ujson

from ..data_utils.almond_utils import (
    ISO_to_LANG,
    detokenize_cjk_chars,
    is_device,
    is_entity,
    is_entity_marker,
    process_id,
    quoted_pattern_with_space,
    tokenize_cjk_chars,
)
from ..data_utils.database import Database
from ..data_utils.example import Example, Feature, get_pad_feature
from ..data_utils.remote_database import RemoteElasticDatabase
from ..model_utils.translation import compute_attention, force_replace_quoted_params, replace_quoted_params
from ..paraphrase.data_utils import input_heuristics, output_heuristics
from .almond_dataset import AlmondDataset
from .base_dataset import Split
from .base_task import BaseTask
from .generic_dataset import CQA, all_tokens_fn, default_batch_fn, input_then_output_len
from .registry import register_task

logger = logging.getLogger(__name__)


class BaseAlmondTask(BaseTask):
    """Base class for the Almond semantic parsing task
    i.e. natural language to formal language (ThingTalk) mapping"""

    def __init__(self, name, args):
        super().__init__(name, args)
        self.args = args
        self.no_feature_fields = ['answer']
        if self.utterance_field() == 'question':
            self.no_feature_fields.append('context')
        else:
            self.no_feature_fields.append('question')

        self._almond_has_multiple_programs = args.almond_has_multiple_programs
        self._almond_detokenize_sentence = args.almond_detokenize_sentence
        self._almond_thingtalk_version = args.almond_thingtalk_version

        self.almond_domains = args.almond_domains
        self.all_schema_types = set()

        if args.do_ned:
            self.unk_id = args.ned_features_default_val[0]
            self.TTtype2qid = dict()
            with open(os.path.join(self.args.database_dir, 'es_material/almond_type2qid.json'), 'r') as fin:
                almond_type2qid = ujson.load(fin)
            for domain in self.almond_domains:
                self.TTtype2qid.update(almond_type2qid[domain])
            self.qid2TTtype = {v: k for k, v in self.TTtype2qid.items()}
            self._init_db()

    def _init_db(self):
        if self.args.database_type == 'json':
            canonical2type = {}
            all_canonicals = marisa_trie.Trie()
            # canonical2type.json is a big file (>4G); load it only when necessary
            if self.args.ned_retrieve_method not in ['bootleg', 'type-oracle']:
                with open(os.path.join(self.args.database_dir, 'es_material/canonical2type.json'), 'r') as fin:
                    canonical2type = ujson.load(fin)
                    all_canonicals = marisa_trie.Trie(canonical2type.keys())
            with open(os.path.join(self.args.database_dir, 'es_material/typeqid2id.json'), 'r') as fin:
                typeqid2id = ujson.load(fin)

            self.db = Database(
                canonical2type, typeqid2id, all_canonicals, self.args.ned_features_default_val, self.args.ned_features_size
            )

        elif self.args.database_type == 'remote-elastic':
            with open(os.path.join(self.args.database_dir, 'es_material/elastic_config.json'), 'r') as fin:
                es_config = ujson.load(fin)
            with open(os.path.join(self.args.database_dir, 'es_material/typeqid2id.json'), 'r') as fin:
                typeqid2id = ujson.load(fin)
            self.db = RemoteElasticDatabase(
                es_config, typeqid2id, self.args.ned_features_default_val, self.args.ned_features_size
            )

    @property
    def utterance_field(self):
        return NotImplementedError

    @property
    def metrics(self):
        return ['em', 'sm', 'f1']

    def _is_program_field(self, field_name):
        raise NotImplementedError()

    def _make_example(self, parts, dir_name, **kwargs):
        raise NotImplementedError()

    def get_splits(self, root, **kwargs):
        return AlmondDataset.return_splits(path=os.path.join(root, 'almond'), make_example=self._make_example, **kwargs)

    def collect_answer_entity_types(self, answer):
        entity2type = dict()

        answer_entities = quoted_pattern_with_space.findall(answer)
        for ent in answer_entities:

            if self._almond_thingtalk_version == 1:
                #  ... param:inAlbum:Entity(org.schema.Music:MusicAlbum) == " XXXX " ...
                # ... param:artists contains " XXX " ^^com.spotify:artist and param:id =~ " XXX " ...
                # ... filter param:geo:Location == location: " XXXX " and ...

                # assume first syntax
                idx = answer.index('" ' + ent + ' "')

                schema_entity_type = None

                answer_tokens_after_entity = answer[idx + len('" ' + ent + ' "') :].split()
                if answer_tokens_after_entity[0].startswith('^^'):
                    schema_entity_type = answer_tokens_after_entity[0].rsplit(':', 1)[1]

                if schema_entity_type is None:

                    answer_tokens_before_entity = answer[:idx].split()

                    # check last three tokens to find one that starts with param
                    for i in range(3):
                        if answer_tokens_before_entity[-i].startswith('param:'):
                            schema_entity_type = answer_tokens_before_entity[-i]
                            break

                    if schema_entity_type:
                        schema_entity_type = schema_entity_type.strip('()').rsplit(':', 1)[1]

                if schema_entity_type is None or schema_entity_type not in self.TTtype2qid.keys():
                    schema_type = self.db.unk_type
                else:
                    schema_type = self.TTtype2qid[schema_entity_type]

                entity2type[ent] = schema_type

            else:

                # ** this should change if thingtalk syntax changes **

                # ( ... [Book|Music|...] ( ) filter id =~ " position and affirm " ) ...'
                # ... ^^org.schema.Book:Person ( " james clavell " ) ...
                # ... contains~ ( [award|genre|...] , " booker " ) ...
                # ... inLanguage =~ " persian " ...

                # missing syntax from current code
                #  ... @com.spotify . song ( ) filter in_array~ ( id , [ " piano man " , " uptown girl " ] ) ) [ 1 : 2 ]  ...

                # assume first syntax
                idx = answer.index('" ' + ent + ' "')

                schema_entity_type = None
                tokens_before_entity = answer[:idx].split()

                if tokens_before_entity[-2] == 'Location':
                    schema_entity_type = 'Location'

                elif tokens_before_entity[-1] in ['>=', '==', '<='] and tokens_before_entity[-2] in [
                    'ratingValue',
                    'reviewCount',
                    'checkoutTime',
                    'checkinTime',
                ]:
                    schema_entity_type = tokens_before_entity[-2]

                elif tokens_before_entity[-1] == '=~':
                    if tokens_before_entity[-2] == 'id':
                        # travers previous tokens until find filter
                        i = -3
                        while tokens_before_entity[i] != 'filter' and i > 3 - len(tokens_before_entity):
                            i -= 1
                        schema_entity_type = tokens_before_entity[i - 3]
                    else:
                        schema_entity_type = tokens_before_entity[-2]

                elif tokens_before_entity[-4] == 'contains~':
                    schema_entity_type = tokens_before_entity[-2]

                elif tokens_before_entity[-2].startswith('^^'):
                    schema_entity_type = tokens_before_entity[-2].rsplit(':', 1)[1]
                    if (
                        schema_entity_type == 'Person'
                        and tokens_before_entity[-3] == 'null'
                        and tokens_before_entity[-5] in ['director', 'creator', 'actor']
                    ):
                        schema_entity_type = 'Person.' + tokens_before_entity[-5]

                if schema_entity_type is None or schema_entity_type not in self.TTtype2qid.keys():
                    schema_type = self.db.unk_type
                else:
                    schema_type = self.TTtype2qid[schema_entity_type]

                entity2type[ent] = schema_type

            self.all_schema_types.add(schema_entity_type)

        return entity2type

    def pad_features(self, features, max_size, pad_id):
        if len(features) > max_size:
            features = features[:max_size]
        else:
            features += [pad_id] * (max_size - len(features))
        return features

    def oracle_type_ids(self, tokens, entity2type):
        tokens_type_ids = [
            [self.args.ned_features_default_val[0]] * self.args.ned_features_size[0] for _ in range(len(tokens))
        ]
        tokens_text = " ".join(tokens)

        for ent, type in entity2type.items():
            ent_num_tokens = len(ent.split(' '))
            if ent in tokens_text:
                idx = tokens_text.index(ent)
            else:
                logger.warning('Found a mismatch between sentence and annotation entities')
                logger.info(f'sentence: {tokens_text}, entity2type: {entity2type}')
                continue
            token_pos = len(tokens_text[:idx].split())

            type_id = self.db.typeqid2id[type]
            type_id = self.pad_features([type_id], self.args.ned_features_size[0], self.args.ned_features_default_val[0])

            tokens_type_ids[token_pos : token_pos + ent_num_tokens] = [type_id] * ent_num_tokens

        return tokens_type_ids

    def find_type_ids(self, tokens, answer):
        tokens_type_ids = []

        if self.args.database_type == 'json':
            if self.args.ned_retrieve_method == 'naive':
                tokens_type_ids = self.db.lookup(
                    tokens, self.args.database_lookup_method, self.args.min_entity_len, self.args.max_entity_len
                )
            elif self.args.ned_retrieve_method == 'entity-oracle':
                answer_entities = quoted_pattern_with_space.findall(answer)
                tokens_type_ids = self.db.lookup(tokens, answer_entities=answer_entities)
            elif self.args.ned_retrieve_method == 'type-oracle':
                entity2type = self.collect_answer_entity_types(answer)
                tokens_type_ids = self.oracle_type_ids(tokens, entity2type)

        return tokens_type_ids

    def find_type_probs(self, tokens, default_val, default_size):
        token_freqs = [[default_val] * default_size] * len(tokens)
        return token_freqs

    def batch_postprocess_prediction_ids(self, batch_example_ids, batch_src_ids, batch_tgt_ids, **kwargs):
        return batch_tgt_ids

    def postprocess_prediction(self, example_id, prediction):

        if self._almond_detokenize_sentence:
            # To make genienlp transparent to the tokenization done by genie-toolkit
            # We tokenize prediction here by adding whitespace between each CJK character
            prediction = tokenize_cjk_chars(prediction)

        new_tokens = []
        for token in prediction.split():
            if token.startswith('STRING_'):
                token = 'QUOTED_' + token
            elif token.startswith('ENTITY_'):
                token = 'GENERIC_' + token
            new_tokens.append(token)

        new_prediction = ' '.join(new_tokens)
        return new_prediction

    def add_type_tokens(self, new_sentence, features, add_types_to_text):
        new_sentence_tokens = new_sentence.split(' ')
        assert len(new_sentence_tokens) == len(features)
        sentence_plus_types_tokens = []
        i = 0
        if add_types_to_text == 'insert':
            while i < len(new_sentence_tokens):
                token = new_sentence_tokens[i]
                feat = features[i]
                # token is an entity
                if any([val != self.args.ned_features_default_val[0] for val in feat.type_id]):
                    final_token = '<e> '
                    all_types = ' | '.join(
                        set(
                            [
                                self.qid2TTtype[self.db.id2type[id]]
                                for id in feat.type_id
                                if self.db.id2type[id] in self.qid2TTtype
                            ]
                        )
                    )
                    final_token += '( ' + all_types + ' ) ' + token
                    # append all entities with same type
                    i += 1
                    while i < len(new_sentence_tokens) and features[i] == feat:
                        final_token += ' ' + new_sentence_tokens[i]
                        i += 1
                    final_token += ' </e>'
                    sentence_plus_types_tokens.append(final_token)
                else:
                    sentence_plus_types_tokens.append(token)
                    i += 1

        elif add_types_to_text == 'append':
            sentence_plus_types_tokens.extend(new_sentence_tokens)
            sentence_plus_types_tokens.append('<e>')
            while i < len(new_sentence_tokens):
                feat = features[i]
                # token is an entity
                if any([val != self.args.ned_features_default_val[0] for val in feat.type_id]):
                    all_types = ' | '.join(
                        set(
                            [
                                self.qid2TTtype[self.db.id2type[id]]
                                for id in feat.type_id
                                if self.db.id2type[id] in self.qid2TTtype
                            ]
                        )
                    )
                    all_tokens = []
                    # append all entities with same type
                    while i < len(new_sentence_tokens) and features[i] == feat:
                        all_tokens.append(new_sentence_tokens[i])
                        i += 1
                    final_token = ' '.join([*all_tokens, '(', all_types, ')', ';'])
                    sentence_plus_types_tokens.append(final_token)
                else:
                    i += 1

            sentence_plus_types_tokens.append('</e>')

        return ' '.join(sentence_plus_types_tokens)

    def preprocess_field(self, sentence, field_name=None, answer=None, preprocess_entities=True):
        if self.override_context is not None and field_name == 'context':
            pad_feature = get_pad_feature(
                self.args.ned_features, self.args.ned_features_default_val, self.args.ned_features_size
            )
            return (
                self.override_context,
                [pad_feature] * len(self.override_context.split(' ')) if pad_feature else [],
                self.override_context,
            )
        if self.override_question is not None and field_name == 'question':
            pad_feature = get_pad_feature(
                self.args.ned_features, self.args.ned_features_default_val, self.args.ned_features_size
            )
            return (
                self.override_question,
                [pad_feature] * len(self.override_question.split(' ')) if pad_feature else [],
                self.override_question,
            )
        if not sentence:
            return '', [], ''

        tokens = sentence.split(' ')
        is_program = self._is_program_field(field_name)
        new_tokens = []
        for token in tokens:
            if (is_entity(token) and preprocess_entities) or (is_program and (is_device(token) or is_entity_marker(token))):
                if token.startswith('QUOTED_STRING_'):
                    token = token[len('QUOTED_') :]
                elif token.startswith('GENERIC_ENTITY_'):
                    token = token[len('GENERIC_') :]

                self.special_tokens.add(token)
            new_tokens.append(token)
        tokens = new_tokens
        new_sentence = ' '.join(tokens)

        if self._almond_detokenize_sentence:

            # BERT tokenizers by default add whitespace around any CJK character
            # SPM-based tokenizers are trained on raw text and do better when receive untokenized text
            # In genienlp we detokenize CJK characters and leave tokenization to the model's tokenizer
            # NOTE: input datasets for almond are usually pretokenized using genie-toolkit which
            # inserts whitespace around any CJK character. This detokenization ensures that SPM-based tokenizers
            # see the text without space between those characters
            new_sentence = detokenize_cjk_chars(new_sentence)
            tokens = new_sentence.split(' ')

            new_sentence = ''
            in_string = False
            for token in tokens:
                if is_program:
                    if token == '"':
                        in_string = not in_string
                    if not in_string:
                        new_sentence += ' ' + token
                        continue
                if token in (',', '.', '?', '!', ':', ')', ']', '}') or token.startswith("'"):
                    new_sentence += token
                else:
                    new_sentence += ' ' + token
        elif is_program and field_name != 'answer':
            new_tokens = []
            in_string = False
            for token in tokens:
                if token == '"':
                    in_string = not in_string
                if in_string:
                    new_tokens.append(token)
                    continue

                if not is_entity(token) and not is_entity_marker(token) and not is_device(token):
                    for word in token.split('_'):
                        new_tokens.append(word)
                else:
                    new_tokens.append(token)
            new_sentence = ' '.join(new_tokens)

        new_sentence = new_sentence.strip()
        new_tokens = new_sentence.split(' ')
        new_sentence_length = len(new_tokens)

        tokens_type_ids, tokens_type_probs = None, None

        if 'type_id' in self.args.ned_features and field_name != 'answer':
            tokens_type_ids = [
                [self.args.ned_features_default_val[0]] * self.args.ned_features_size[0] for _ in range(new_sentence_length)
            ]
        if 'type_prob' in self.args.ned_features and field_name != 'answer':
            tokens_type_probs = [
                [self.args.ned_features_default_val[1]] * self.args.ned_features_size[1] for _ in range(new_sentence_length)
            ]

        if self.args.do_ned and self.args.ned_retrieve_method != 'bootleg' and field_name not in self.no_feature_fields:
            if 'type_id' in self.args.ned_features:
                tokens_type_ids = self.find_type_ids(new_tokens, answer)
            if 'type_prob' in self.args.ned_features:
                tokens_type_probs = self.find_type_probs(
                    new_tokens, self.args.ned_features_default_val[1], self.args.ned_features_size[1]
                )

            if self.args.verbose and self.args.do_ned:
                print()
                print(
                    *[f'token: {token}\ttype: {token_type}' for token, token_type in zip(new_tokens, tokens_type_ids)],
                    sep='\n',
                )

        zip_list = []
        if tokens_type_ids:
            assert len(tokens_type_ids) == new_sentence_length
            zip_list.append(tokens_type_ids)
        if tokens_type_probs:
            assert len(tokens_type_probs) == new_sentence_length
            zip_list.append(tokens_type_probs)

        features = [Feature(*tup) for tup in zip(*zip_list)]

        sentence_plus_types = new_sentence
        if self.args.do_ned and self.args.add_types_to_text != 'no' and len(features) and field_name == self.utterance_field():
            sentence_plus_types = self.add_type_tokens(new_sentence, features, self.args.add_types_to_text)

        return new_sentence, features, sentence_plus_types


@register_task('almond')
class Almond(BaseAlmondTask):
    """The Almond semantic parsing task
    i.e. natural language to formal language (ThingTalk) mapping"""

    def _is_program_field(self, field_name):
        return field_name == 'answer'

    def utterance_field(self):
        return 'context'

    def _make_example(self, parts, dir_name=None, **kwargs):
        # the question is irrelevant, so the question says English and ThingTalk even if we're doing
        # a different language (like Chinese)
        if self._almond_has_multiple_programs:
            example_id, sentence, target_code = parts[:3]
        else:
            example_id, sentence, target_code = parts
        question = 'translate from english to thingtalk'
        context = sentence
        answer = target_code
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )


@register_task('almond_natural_seq2seq')
class NaturalSeq2Seq(BaseAlmondTask):
    """
    The Almond sequence to sequence task where both sequences are natural language.
    In this task entities (see ENTITY_REGEX) are not preprocessed in contrast to paraphrasing and translation tasks.
    Paraphrasing and translation inherit from this class.
    """

    @property
    def metrics(self):
        return ['bleu', 'em', 'nf1']

    def _is_program_field(self, field_name):
        return False

    def utterance_field(self):
        return 'context'

    def _make_example(self, parts, dir_name=None, **kwargs):
        # the question is irrelevant
        if len(parts) == 2:
            input_sequence, target_sequence = parts
            example_id = "id-null"
        elif len(parts) == 3:
            example_id, input_sequence, target_sequence = parts
        else:
            raise ValueError(f'Input file contains line with {len(parts)} parts: {str(parts)}')
        question = 'translate from input to output'
        context = input_sequence
        answer = target_sequence
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def preprocess_field(self, sentence, field_name=None, answer=None, preprocess_entities=False):
        return super().preprocess_field(sentence, field_name, answer, preprocess_entities=False)

    def get_splits(self, root, **kwargs):
        return AlmondDataset.return_splits(path=os.path.join(root, 'almond'), make_example=self._make_example, **kwargs)


@register_task('almond_paraphrase')
class Paraphrase(NaturalSeq2Seq):
    """The Almond paraphrasing task. Applies the necessary preprocessing for special tokens and case changes.
    Can be used at prediction and training time. Training is still experimental.
    """

    def __init__(self, name, args):
        super().__init__(name, args)
        self.reverse_maps = {}

    @property
    def metrics(self):
        return ['bleu']

    def postprocess_prediction(self, example_id, prediction):
        return output_heuristics(prediction, self.reverse_maps[example_id])

    def _make_example(self, parts, dir_name=None, **kwargs):
        if len(parts) == 3:
            example_id, sentence, thingtalk = parts
        elif len(parts) == 4:
            example_id, _, sentence, thingtalk = parts  # ignore dialogue context
        else:
            raise ValueError(f'Input file contains line with {len(parts)} parts: {str(parts)}')

        example_id = self.name + '/' + example_id

        sentence, reverse_map = input_heuristics(sentence, thingtalk=thingtalk, is_cased=True)
        # this task especially needs example ids to be unique
        while example_id in self.reverse_maps:
            example_id += '.'
        self.reverse_maps[example_id] = reverse_map

        question = 'translate from input to output'
        context = sentence
        answer = sentence  # means we calculate self-bleu

        return Example.from_raw(example_id, context, question, answer, preprocess=self.preprocess_field, lower=False)


@register_task('almond_translate')
class Translate(NaturalSeq2Seq):
    """
    Almond translation task: Translate a sentence from one language to another.
    Can be used at prediction and training time. Training is still experimental.
    """

    def __init__(self, name, args):
        super().__init__(name, args)

    @property
    def metrics(self):
        return ['bleu']

    def postprocess_prediction(self, example_id, prediction):
        return super().postprocess_prediction(example_id, prediction)

    def batch_postprocess_prediction_ids(self, batch_example_ids, batch_src_ids, batch_tgt_ids, **kwargs):

        numericalizer = kwargs.pop('numericalizer')
        cross_attentions = kwargs.pop('cross_attentions')

        # TODO _tokenizer should not be private
        tokenizer = numericalizer._tokenizer

        all_src_tokens = numericalizer.convert_ids_to_tokens(batch_src_ids, skip_special_tokens=False)
        all_tgt_tokens = numericalizer.convert_ids_to_tokens(batch_tgt_ids, skip_special_tokens=False)

        # remove language code from the beginning of src_tokens and shift layer_attention
        len_prefix_wp = len(tokenizer.tokenize(numericalizer.input_prefix))
        all_src_tokens = [tokens[len_prefix_wp:] for tokens in all_src_tokens]
        cross_attentions = cross_attentions[:, :, :, len_prefix_wp:]

        cross_attention_pooled = compute_attention(cross_attentions, att_pooling=self.args.att_pooling, dim=1)

        all_text_outputs = []
        # post-process predictions ids
        for i, (src_tokens, tgt_tokens, cross_att) in enumerate(zip(all_src_tokens, all_tgt_tokens, cross_attention_pooled)):

            # shift target tokens left to match the attention positions
            if tgt_tokens[0] in tokenizer.all_special_tokens:
                tgt_tokens = tgt_tokens[1:]

            # remove all trailing special tokens from source
            while src_tokens[-1] in tokenizer.all_special_tokens:
                src_tokens = src_tokens[:-1]

            # crop to match src and tgt new lengths
            cross_att = cross_att[: len(tgt_tokens), : len(src_tokens)]

            # plot cross-attention heatmap
            if self.args.plot_heatmaps:
                import matplotlib.pyplot as plt
                import seaborn as sns

                graph = sns.heatmap(torch.log(cross_att), xticklabels=src_tokens, yticklabels=tgt_tokens)
                graph.set_xticklabels(graph.get_xmajorticklabels(), fontsize=12)
                graph.set_yticklabels(graph.get_ymajorticklabels(), fontsize=12)

                plt.savefig(os.path.join(os.path.dirname(self.args.save), f'heatmap_{batch_example_ids[i]}'))
                plt.show()

            # remove eos and all pad tokens if present
            while tgt_tokens[-1] in tokenizer.all_special_tokens:
                tgt_tokens = tgt_tokens[:-1]

            if self.args.replace_qp:
                text, is_replaced = replace_quoted_params(src_tokens, tgt_tokens, tokenizer, cross_att)
                if not is_replaced and self.args.force_replace_qp:
                    text = force_replace_quoted_params(src_tokens, tgt_tokens, tokenizer, cross_att)
            else:
                text = tokenizer.convert_tokens_to_string(tgt_tokens)

            all_text_outputs.append(text)

            with tokenizer.as_target_tokenizer():
                partial_batch_prediction_ids = tokenizer.batch_encode_plus(
                    all_text_outputs, padding=True, return_tensors='pt'
                )['input_ids']

        return partial_batch_prediction_ids

    def _make_example(self, parts, dir_name=None, **kwargs):
        # answer has to be provided by default unless doing prediction
        no_answer = kwargs.get('translate_no_answer', False)
        example_id = 'id-null'
        if not no_answer:
            if len(parts) == 2:
                sentence, answer = parts
            elif len(parts) == 3:
                example_id, sentence, answer = parts
            elif len(parts) == 4:
                example_id, sentence, answer, thingtalk = parts
            elif len(parts) == 5:
                example_id, _, sentence, answer, thingtalk = parts  # ignore dialogue context
            else:
                raise ValueError(f'Input file contains line with {len(parts)} parts: {str(parts)}')
        else:
            if len(parts) == 1:
                sentence = parts
            elif len(parts) == 2:
                example_id, sentence = parts
            elif len(parts) == 3:
                example_id, sentence, thingtalk = parts
            elif len(parts) == 4:
                example_id, _, sentence, thingtalk = parts  # ignore dialogue context
            else:
                raise ValueError(f'Input file contains line with {len(parts)} parts: {str(parts)}')

        question = 'translate from input to output'
        context = sentence

        # no answer is provided
        if no_answer:
            answer = '.'

        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )


@register_task('contextual_almond')
class ContextualAlmond(BaseAlmondTask):
    """Contextual Almond semantic parsing task"""

    def _is_program_field(self, field_name):
        return field_name in ('answer', 'context')

    def utterance_field(self):
        return 'question'

    def _make_example(self, parts, dir_name=None, **kwargs):
        if self._almond_has_multiple_programs:
            example_id, context, sentence, target_code = parts[:4]
        else:
            example_id, context, sentence, target_code = parts
        answer = target_code
        question = sentence
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )


@register_task('reverse_almond')
class ReverseAlmond(BaseAlmondTask):
    """Reverse Almond semantic parsing task
    i.e. formal language to natural language mapping"""

    @property
    def metrics(self):
        return ['bleu', 'em']

    def utterance_field(self):
        return 'context'

    def _is_program_field(self, field_name):
        return field_name == 'context'

    def _make_example(self, parts, dir_name=None, **kwargs):
        # the question is irrelevant, so the question says English and ThingTalk even if we're doing
        # a different language (like Chinese)
        example_id, sentence, target_code = parts
        question = 'translate from thingtalk to english'
        context = target_code
        answer = sentence
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )


# TODO add a similar preprocessing step to Multilingual dialogue tasks as well
class BaseAlmondDialogueNLUTask(BaseAlmondTask):
    def preprocess_field(self, sentence, field_name=None, answer=None):
        if not sentence:
            return sentence, [], sentence

        # remove the $dialogue at the start of the dialogue
        # this is safe because we know we're processing dialogues, so the answer
        # always starts with $dialogue and the context is either `null` or also
        # starts with $dialogue
        if field_name in ['context', 'answer'] and sentence.startswith('$dialogue '):
            sentence = sentence[len('$dialogue ') :]
        return super().preprocess_field(sentence, field_name, answer)

    def postprocess_prediction(self, example_id, prediction):
        prediction = super().postprocess_prediction(example_id, prediction)
        if not prediction.startswith('$'):
            return '$dialogue ' + prediction
        return prediction


@register_task('almond_dialogue_nlu')
class AlmondDialogueNLU(BaseAlmondDialogueNLUTask):
    """Multi-turn NLU task for Almond dialogues
    (translate the user utterance to a formal representation, given the current
    state of the conversation)
    """

    def _is_program_field(self, field_name):
        return field_name in ('answer', 'context')

    def utterance_field(self):
        return 'question'

    def _make_example(self, parts, dir_name=None, **kwargs):
        if self._almond_has_multiple_programs:
            example_id, context, sentence, target_code = parts[:4]
        else:
            example_id, context, sentence, target_code = parts

        answer = target_code
        question = sentence
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        return AlmondDataset.return_splits(path=os.path.join(root, 'almond/user'), make_example=self._make_example, **kwargs)


@register_task('almond_dialogue_nlu_agent')
class AlmondDialogueNLUAgent(BaseAlmondDialogueNLUTask):
    """Multi-turn NLU task for Almond dialogues, for the agent utterance
    (translate the agent utterance to a formal representation, given the current
    state of the conversation).
    This is used to facilitate annotation of human-human dialogues.
    """

    def _is_program_field(self, field_name):
        return field_name in ('answer', 'context')

    def utterance_field(self):
        return 'question'

    def _make_example(self, parts, dir_name=None, **kwargs):
        if self._almond_has_multiple_programs:
            example_id, context, sentence, target_code = parts[:4]
        else:
            example_id, context, sentence, target_code = parts
        answer = target_code
        question = sentence
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        return AlmondDataset.return_splits(path=os.path.join(root, 'almond/agent'), make_example=self._make_example, **kwargs)


@register_task('almond_dialogue_nlg')
class AlmondDialogueNLG(BaseAlmondTask):
    """Multi-turn NLG task for Almond dialogues
    (generate the system utterance, given the current state of the conversation
    and the desired system dialogue act)
    """

    def _is_program_field(self, field_name):
        return field_name in ('context', 'question')

    def utterance_field(self):
        return 'answer'

    @property
    def metrics(self):
        return ['bleu']

    def _make_example(self, parts, dir_name=None, **kwargs):
        # the question is irrelevant for this task
        example_id, context, sentence, target_code = parts
        question = target_code
        answer = sentence
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        return AlmondDataset.return_splits(path=os.path.join(root, 'almond/nlg'), make_example=self._make_example, **kwargs)


@register_task('almond_dialogue_policy')
class AlmondDialoguePolicy(BaseAlmondTask):
    """Multi-turn dialogue policy task for Almond dialogues
    (generate the next dialogue act, given the current state of the conversation)
    """

    def _is_program_field(self, field_name):
        return field_name in ('answer', 'context')

    def utterance_field(self):
        return 'question'

    @property
    def metrics(self):
        return ['em', 'f1']

    def _make_example(self, parts, dir_name=None, **kwargs):
        # the question is irrelevant for this task, and the sentence is intentionally ignored
        example_id, context, _sentence, target_code = parts
        question = 'what should the agent do ?'
        answer = target_code
        return Example.from_raw(
            self.name + '/' + example_id, context, question, answer, preprocess=self.preprocess_field, lower=False
        )

    def get_splits(self, root, **kwargs):
        return AlmondDataset.return_splits(path=os.path.join(root, 'almond/agent'), make_example=self._make_example, **kwargs)


class BaseAlmondMultiLingualTask(BaseAlmondTask):
    """Base task for MultiLingual Almond"""

    def get_train_processed_ids(self, split):
        all_ids = []
        for ex in split.examples:
            all_ids.append(process_id(ex))
        return all_ids

    def combine_datasets(self, datasets, all_paths, sort_key_fn, batch_size_fn, used_fields, groups):
        splits = defaultdict()
        # paths = defaultdict()

        for field in used_fields:
            # choose one path and replace dir name with 'combined'
            # paths[field] = '/'.join([getattr(all_paths[0], field).rsplit('/', 2)[0], 'combined', getattr(all_paths[0], field).rsplit('/', 2)[2]])

            all_examples = []
            for dataset in datasets:
                all_examples.extend(getattr(dataset, field).examples)

            splits[field] = CQA(all_examples, sort_key_fn=sort_key_fn, batch_size_fn=batch_size_fn, groups=groups)

        return Split(train=splits.get('train'), eval=splits.get('eval'), test=splits.get('test'), aux=splits.get('aux'))
        # Split(train=paths.get('train'), eval=paths.get('eval'), test=paths.get('test'), aux=paths.get('aux'))

    def get_splits(self, root, **kwargs):
        all_datasets = []
        all_paths = []
        # number of directories to read data from
        all_dirs = kwargs['all_dirs'].split('+')

        for dir in all_dirs:
            splits, paths = AlmondDataset.return_splits(
                path=os.path.join(root, 'almond/{}'.format(dir)), make_example=self._make_example, **kwargs
            )
            all_datasets.append(splits)
            all_paths.append(paths)

        used_fields = [field for field in all_datasets[0]._fields if getattr(all_datasets[0], field) is not None]

        assert len(all_datasets) >= 1
        if kwargs.get('sentence_batching'):
            for field in used_fields:
                lengths = list(map(lambda dataset: len(getattr(dataset, field)), all_datasets))
                assert len(set(lengths)) == 1, 'When using sentence batching your datasets should have the same size.'
            if 'train' in used_fields:
                ids_sets = list(map(lambda dataset: set(self.get_train_processed_ids(dataset.train)), all_datasets))
                id_set_base = set(ids_sets[0])
                for id_set in ids_sets:
                    assert set(id_set) == id_set_base, 'When using sentence batching your datasets should have matching ids'

            sort_key_fn = process_id
            batch_size_fn = default_batch_fn
        else:
            # use default values for `sort_key_fn` and `batch_size_fn`
            sort_key_fn = input_then_output_len
            batch_size_fn = all_tokens_fn

        groups = len(all_datasets) if kwargs.get('sentence_batching') else None

        if kwargs.get('separate_eval') and (all_datasets[0].eval or all_datasets[0].test):
            return all_datasets, all_paths
        # TODO fix handling paths for multilingual
        else:
            return (
                self.combine_datasets(all_datasets, all_paths, sort_key_fn, batch_size_fn, used_fields, groups),
                all_paths[0],
            )


@register_task('almond_multilingual')
class AlmondMultiLingual(BaseAlmondMultiLingualTask):
    def __init__(self, name, args):
        super().__init__(name, args)
        self.lang_as_question = args.almond_lang_as_question

    def _is_program_field(self, field_name):
        return field_name == 'answer'

    def utterance_field(self):
        return 'context'

    @property
    def metrics(self):
        return ['em', 'sm', 'bleu']

    def _make_example(self, parts, dir_name, **kwargs):
        if self._almond_has_multiple_programs:
            example_id, sentence, target_code = parts[:3]
        else:
            example_id, sentence, target_code = parts
        language = ISO_to_LANG.get(dir_name, 'English').lower()
        if self.lang_as_question:
            question = 'translate from {} to thingtalk'.format(language)
        else:
            question = 'translate from english to thingtalk'
        context = sentence
        answer = target_code
        return Example.from_raw(
            self.name + '/' + dir_name + '/' + example_id,
            context,
            question,
            answer,
            preprocess=self.preprocess_field,
            lower=False,
        )


@register_task('almond_dialogue_multilingual_nlu')
class AlmondDialogMultiLingualNLU(BaseAlmondMultiLingualTask):
    """Multi-turn NLU task (user and agent) for MultiLingual Almond dialogues"""

    def _is_program_field(self, field_name):
        return field_name in ('answer', 'context')

    def utterance_field(self):
        return 'question'

    @property
    def metrics(self):
        return ['em', 'sm', 'bleu']

    def _make_example(self, parts, dir_name=None, **kwargs):
        if self._almond_has_multiple_programs:
            example_id, context, sentence, target_code = parts
        else:
            example_id, context, sentence, target_code = parts[:4]
        answer = target_code
        question = sentence
        return Example.from_raw(
            self.name + '/' + dir_name + '/' + example_id,
            context,
            question,
            answer,
            preprocess=self.preprocess_field,
            lower=False,
        )


@register_task('almond_dialogue_multilingual_nlg')
class AlmondDialogMultiLingualNLG(BaseAlmondTask):
    """Multi-turn NLG task (agent) for MultiLingual Almond dialogues"""

    def _is_program_field(self, field_name):
        return field_name == 'context'

    def utterance_field(self):
        return 'question'

    @property
    def metrics(self):
        return ['bleu']

    def _make_example(self, parts, dir_name=None, **kwargs):
        # the question is irrelevant for this task
        example_id, context, sentence, target_code = parts
        question = 'what should the agent say ?'
        context = context + ' ' + target_code
        answer = sentence
        return Example.from_raw(
            self.name + '/' + dir_name + '/' + example_id,
            context,
            question,
            answer,
            preprocess=self.preprocess_field,
            lower=False,
        )



import logging
import random

import torch
import torch.nn as nn
from transformers import XLMRobertaConfig
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions
from transformers.models.bert.modeling_bert import BertEmbeddings, BertModel
from transformers.models.gpt2 import tokenization_gpt2
from transformers.models.marian.convert_marian_tatoeba_to_pytorch import GROUP_MEMBERS
from transformers.models.marian.convert_marian_to_pytorch import GROUPS
from transformers.models.mbart.tokenization_mbart import MBartTokenizer
from transformers.models.roberta.modeling_roberta import (
    RobertaEncoder,
    RobertaPooler,
    RobertaPreTrainedModel,
    create_position_ids_from_input_ids,
)
from transformers.models.t5 import tokenization_t5

logger = logging.getLogger(__name__)


BART_PRETRAINED_CONFIG_ARCHIVE_MAP = {
    # official models
    "facebook/bart-base": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/bart-base/config.json",
    "facebook/bart-large": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/bart-large/config.json",
    "facebook/bart-large-mnli": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/bart-large-mnli/config.json",
    "facebook/bart-large-cnn": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/bart-large-cnn/config.json",
    "facebook/bart-large-xsum": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/bart-large-xsum/config.json",
    # community models; see https://huggingface.co/models?filter=bart for more
    "sshleifer/bart-tiny-random": "https://s3.amazonaws.com/models.huggingface.co/bert/sshleifer/bart-tiny-random/config.json",
}

MBART_PRETRAINED_CONFIG_ARCHIVE_MAP = {
    # official models
    "facebook/mbart-large-en-ro": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/mbart-large-en-ro/config.json",
    "facebook/mbart-large-cc25": "https://s3.amazonaws.com/models.huggingface.co/bert/facebook/mbart-large-cc25/config.json",
    # community models; see https://huggingface.co/models?filter=mbart for more
    "sshleifer/tiny-mbart": "https://s3.amazonaws.com/models.huggingface.co/bert/sshleifer/tiny-mbart/config.json",
}

MT5_PRETRAINED_CONFIG_ARCHIVE_MAP = {
    'google/mt5-{}'.format(v): "https://s3.amazonaws.com/models.huggingface.co/bert/google/mt5-{}/config.json".format(v)
    for v in ['small', 'base', 'large', 'xl', 'xxl']
}


BART_MODEL_LIST = list(BART_PRETRAINED_CONFIG_ARCHIVE_MAP.keys())
MBART_MODEL_LIST = list(MBART_PRETRAINED_CONFIG_ARCHIVE_MAP.keys())
MT5_MODEL_LIST = list(MT5_PRETRAINED_CONFIG_ARCHIVE_MAP.keys())


# all MarianMT models use the same config
MARIAN_PRETRAINED_CONFIG_ARCHIVE_MAP = {
    "Helsinki-NLP/opus-mt-en-de": "https://s3.amazonaws.com/models.huggingface.co/bert/Helsinki-NLP/opus-mt-en-de/config.json",
}


MARIAN_GROUPS = {item[1]: set(item[0].split('+')) for item in GROUPS}
MARIAN_TATOEBA_GROUPS = {k: set(v[1]) for k, v in GROUP_MEMBERS.items()}

MARIAN_GROUP_MEMBERS = {**MARIAN_GROUPS, **MARIAN_TATOEBA_GROUPS}


MODEL_PARALLEL_SUPPORTED_MODELS = (
    list(tokenization_gpt2.PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES.keys())
    + list(tokenization_t5.PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES.keys())
    + list(MT5_PRETRAINED_CONFIG_ARCHIVE_MAP.keys())
)

###############


# TODO try fixing this upstream
class GenieMBartTokenizer(MBartTokenizer):
    '''
    MBartTokenizer with the temporary fix for off-by-one error during generation: https://github.com/huggingface/transformers/issues/5755
    '''

    def __init__(self, *args, tokenizer_file=None, **kwargs):
        super().__init__(*args, tokenizer_file=tokenizer_file, **kwargs)

    def set_src_lang_special_tokens(self, src_lang) -> None:
        """Reset the special tokens to the source lang setting. Prefix [bos_token_id], suffix =[eos_token_id]."""
        self.cur_lang_code = self.lang_code_to_id[src_lang]
        self.prefix_tokens = [self.bos_token_id]
        self.suffix_tokens = [self.eos_token_id]

    def set_tgt_lang_special_tokens(self, lang: str) -> None:
        """Reset the special tokens to the target language setting. Prefix [tgt_lang_code], suffix =[eos_token_id]."""
        self.cur_lang_code = self.lang_code_to_id[lang]
        self.prefix_tokens = [self.cur_lang_code]
        self.suffix_tokens = [self.eos_token_id]


###############
## NER models
###############
## BERT


class BertEmbeddingsForNER(BertEmbeddings):
    """Construct the embeddings from word, position, token_type, and entity_type embeddings."""

    def __init__(self, config, num_db_types, db_unk_id):
        super().__init__(config)
        self.num_db_types = num_db_types
        self.db_unk_id = db_unk_id
        self.pad_token_id = config.pad_token_id
        if num_db_types > 0:
            self.entity_type_embeddings = nn.Embedding(num_db_types, config.hidden_size, padding_idx=db_unk_id)

    def forward(
        self,
        input_ids=None,
        token_type_ids=None,
        position_ids=None,
        inputs_embeds=None,
        entity_ids=None,
        entity_masking=None,
        entity_probs=None,
        entity_word_embeds_dropout=False,
    ):
        if input_ids is not None:
            input_shape = input_ids.size()
        else:
            input_shape = inputs_embeds.size()[:-1]

        seq_length = input_shape[1]

        if position_ids is None:
            position_ids = self.position_ids[:, :seq_length]

        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=self.position_ids.device)

        # dropout word embeddings for entities
        if entity_masking is not None and entity_word_embeds_dropout > 0.0:
            if entity_word_embeds_dropout > random.random():
                input_ids = input_ids * ~(entity_masking.max(-1)[0].bool())

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)

        token_type_embeddings = self.token_type_embeddings(token_type_ids)

        embeddings = inputs_embeds + token_type_embeddings

        if self.position_embedding_type == "absolute":
            position_embeddings = self.position_embeddings(position_ids)
            embeddings += position_embeddings

        if self.num_db_types > 0 and entity_ids is not None:
            # get length for unpadded types
            type_lengths = entity_masking.sum(-1)

            # avoid division by zero
            type_lengths[type_lengths == 0] = 1

            entity_type_embeddings = self.entity_type_embeddings(entity_ids)

            # weighted average
            if entity_probs is not None:
                entity_type_embeddings = entity_type_embeddings * entity_probs.unsqueeze(-1)

            # average embedding of different types
            # size (batch, length, num_types, emb_dim)
            entity_type_embeddings = entity_type_embeddings.sum(-2) / type_lengths.unsqueeze(-1)

            embeddings += entity_type_embeddings

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class BertModelForNER(BertModel):
    """
    Subcalss of BertModel model with an additional entity type embedding layer in the bottom
    """

    def __init__(self, config, num_db_types, db_unk_id):
        super().__init__(config)

        self.embeddings = BertEmbeddingsForNER(config, num_db_types, db_unk_id)
        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        entity_ids=None,
        entity_masking=None,
        entity_probs=None,
        entity_word_embeds_dropout=False,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        embedding_output=None,
    ):

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if attention_mask is None:
            attention_mask = torch.ones(input_shape, device=device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape, device)

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if self.config.is_decoder and encoder_hidden_states is not None:
            encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
            if encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
            encoder_extended_attention_mask = self.invert_attention_mask(encoder_attention_mask)
        else:
            encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        if embedding_output is None:
            embedding_output = self.embeddings(
                input_ids=input_ids,
                position_ids=position_ids,
                token_type_ids=token_type_ids,
                inputs_embeds=inputs_embeds,
                entity_ids=entity_ids,
                entity_masking=entity_masking,
                entity_probs=entity_probs,
                entity_word_embeds_dropout=entity_word_embeds_dropout,
            )

        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


###############
## Roberta


class RobertaEmbeddingsForNER(BertEmbeddingsForNER):
    """
    Same as BertEmbeddings with a tiny tweak for positional embeddings indexing and adding entity_types.
    """

    def __init__(self, config, num_db_types, db_unk_id):
        super().__init__(config, num_db_types, db_unk_id)
        self.padding_idx = config.pad_token_id
        self.position_embeddings = nn.Embedding(
            config.max_position_embeddings, config.hidden_size, padding_idx=self.padding_idx
        )

    def forward(
        self,
        input_ids=None,
        token_type_ids=None,
        position_ids=None,
        inputs_embeds=None,
        past_key_values_length=0,
        entity_ids=None,
        entity_masking=None,
        entity_probs=None,
        entity_word_embeds_dropout=False,
    ):
        if position_ids is None:
            if input_ids is not None:
                # Create the position ids from the input token ids. Any padded tokens remain padded.
                position_ids = create_position_ids_from_input_ids(input_ids, self.padding_idx).to(input_ids.device)
            else:
                position_ids = self.create_position_ids_from_inputs_embeds(inputs_embeds)

        return super().forward(
            input_ids,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
            entity_ids=entity_ids,
            entity_masking=entity_masking,
            entity_probs=entity_probs,
            entity_word_embeds_dropout=entity_word_embeds_dropout,
        )

    def create_position_ids_from_inputs_embeds(self, inputs_embeds):
        """
        We are provided embeddings directly. We cannot infer which are padded so just generate sequential position ids.

        Args:
            inputs_embeds: torch.Tensor

        Returns: torch.Tensor
        """
        input_shape = inputs_embeds.size()[:-1]
        sequence_length = input_shape[1]

        position_ids = torch.arange(
            self.padding_idx + 1, sequence_length + self.padding_idx + 1, dtype=torch.long, device=inputs_embeds.device
        )
        return position_ids.unsqueeze(0).expand(input_shape)


class RobertaModelForNER(RobertaPreTrainedModel):
    def __init__(self, config, num_db_types, db_unk_id, add_pooling_layer=True):
        super().__init__(config)
        self.config = config

        self.embeddings = RobertaEmbeddingsForNER(config, num_db_types, db_unk_id)
        self.encoder = RobertaEncoder(config)

        self.pooler = RobertaPooler(config) if add_pooling_layer else None

        self.init_weights()

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        entity_ids=None,
        entity_masking=None,
        entity_probs=None,
        entity_word_embeds_dropout=False,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        embedding_output=None,
    ):

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        device = input_ids.device if input_ids is not None else inputs_embeds.device

        if attention_mask is None:
            attention_mask = torch.ones(input_shape, device=device)
        if token_type_ids is None:
            token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape, device)

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if self.config.is_decoder and encoder_hidden_states is not None:
            encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
            if encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
            encoder_extended_attention_mask = self.invert_attention_mask(encoder_attention_mask)
        else:
            encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        if embedding_output is None:
            embedding_output = self.embeddings(
                input_ids=input_ids,
                position_ids=position_ids,
                token_type_ids=token_type_ids,
                inputs_embeds=inputs_embeds,
                entity_ids=entity_ids,
                entity_masking=entity_masking,
                entity_probs=entity_probs,
                entity_word_embeds_dropout=entity_word_embeds_dropout,
            )

        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


class XLMRobertaModelForNER(RobertaModelForNER):
    """
    Subcalss of XLMRobertaModel model with an additional entity type embedding layer at the bottom
    """

    config_class = XLMRobertaConfig
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def __init__(self, config, num_db_types, db_unk_id, add_pooling_layer=True):
        super().__init__(config, num_db_types, db_unk_id, add_pooling_layer)



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import itertools
import logging
import os
from typing import Callable, Iterable, List, Tuple, Union

import dill
import numpy as np
import torch
import xgboost as xgb
from sklearn.metrics import accuracy_score, auc, confusion_matrix, precision_recall_curve
from sklearn.model_selection import train_test_split

from .util import ConfidenceFeatures

logger = logging.getLogger(__name__)


# Feature function builders
# drop means after applying dropout, nodrop means before


def max_of(f: Callable) -> Callable:
    return lambda x: f(x).max().view(-1)


def min_of(f: Callable) -> Callable:
    return lambda x: f(x).min().view(-1)


def neg_of(f: Callable) -> Callable:
    return lambda x: -f(x)


def cv_drop_logit(i: int) -> Callable:
    def f(x):
        a = torch.sqrt(var_drop_logit(i)(x)) / mean_drop_logit(i)(x)
        a[a.isnan()] = 0
        a[a.isinf()] = 0
        return a

    return f


def var_drop_logit(i: int) -> Callable:
    return lambda x: torch.var(x[i].drop_logits, dim=0).view(-1)


def var_drop_top2_probs(i: int) -> Callable:
    return lambda x: torch.var(x[i].drop_top2_probs, dim=0).view(-1)


def probability_that_2_overtakes_1(i):
    return lambda x: (x[i].drop_top2_probs > x[i].drop_top1_probs).float().mean(dim=0).view(-1)


def diff_mean_drop_probability_2_and_1(i):
    return lambda x: (x[i].drop_top2_probs.mean(dim=0).float() - x[i].drop_top1_probs.mean(dim=0).float()).view(-1)


def diff_var_drop_probability_2_and_1(i):
    return lambda x: (x[i].drop_top2_probs.var(dim=0).float() - x[i].drop_top1_probs.var(dim=0).float()).view(-1)


def diff_nodrop_probability_2_and_1(i):
    return lambda x: (x[i].nodrop_top2_probs - x[i].nodrop_top1_probs).view(-1)


def mean_drop_logit(i: int) -> Callable:
    return lambda x: torch.mean(x[i].drop_logits, dim=0).view(-1)


def nodrop_entropies(i: int) -> Callable:
    return lambda x: x[i].nodrop_entropies


def nodrop_logit(i: int) -> Callable:
    return lambda x: x[i].nodrop_logits


def prediction_length(i: int) -> Callable:
    return lambda x: torch.tensor(len(x[i].nodrop_logits)).view(-1)


def input_length(i: int) -> Callable:
    return lambda x: torch.tensor(len(x[i].context)).view(-1)


def nodrop_avg_logprob(i: int):
    return lambda x: torch.mean(x[i].nodrop_logits).view(-1)


def variance_of_beam_logits(x):
    a = torch.var(torch.tensor([torch.mean(x[i].nodrop_logits).item() for i in range(1, 5)])).view(-1)
    return a


def variance_of_beam_probs(x):
    a = torch.var(torch.tensor([torch.mean(x[i].nodrop_probs).item() for i in range(1, 5)])).view(-1)
    return a


def mean_drop_avg_logprob(i):
    return lambda x: torch.mean(x[i].drop_logits).view(-1)


def var_drop_avg_logprob(i):
    return lambda x: torch.var(torch.mean(x[i].drop_logits, dim=1)).view(-1)


def cv_drop_avg_logprob(i):
    def f(x):
        a = torch.sqrt(var_drop_avg_logprob(i)(x)) / mean_drop_avg_logprob(i)(x)
        a[a.isnan()] = 0
        a[a.isinf()] = 0
        return a

    return f


def nodrop_prob(i):
    return lambda x: x[i].nodrop_probs.view(-1)


def nodrop_seq_prob(i):
    return lambda x: torch.prod(x[i].nodrop_probs).view(-1)


def mean_drop_seq_prob(i):
    return lambda x: torch.mean(torch.prod(x[i].drop_probs, dim=1)).view(-1)


def prob_first_mistake(i):
    """
    probability that this is the first mistake
    """

    def f(x):
        probs = mean_drop_prob(i)(x)
        ret = torch.zeros_like(probs)
        for j in range(len(probs)):
            ret[j] = torch.prod(probs[:j]) * (1 - probs[j])
        return ret

    return f


def mean_drop_prob(i):
    return lambda x: torch.mean(x[i].drop_probs, dim=0).view(-1)


def var_drop_seq_prob(i):
    return lambda x: torch.var(torch.prod(x[i].drop_probs, dim=1)).view(-1)


def var_drop_prob(i):
    return lambda x: torch.var(x[i].drop_probs, dim=0).view(-1)


def cv_drop_seq_prob(i):
    def f(x):
        a = torch.sqrt(var_drop_seq_prob(i)(x)) / mean_drop_seq_prob(i)(x)
        a[a.isnan()] = 0
        a[a.isinf()] = 0
        return a

    return f


def cv_drop_prob(i: int) -> Callable:
    def f(x):
        a = torch.sqrt(var_drop_prob(i)(x)) / mean_drop_prob(i)(x)
        a[a.isnan()] = 0
        a[a.isinf()] = 0
        return a

    return f


def cev_drop_seq_prob(i):
    """
    Introduced in https://arxiv.org/pdf/1909.00157.pdf
    """

    def f(x):
        a = torch.square(1 - (var_drop_seq_prob(i)(x) / mean_drop_seq_prob(i)(x)))
        a[a.isnan()] = 0
        a[a.isinf()] = 0
        return a

    return f


def cev_drop_prob(i):
    """
    Introduced in https://arxiv.org/pdf/1909.00157.pdf
    """

    def f(x):
        a = torch.square(1 - (var_drop_prob(i)(x) / mean_drop_prob(i)(x)))
        a[a.isnan()] = 0
        a[a.isinf()] = 0
        return a

    return f


def accuracy_at_pass_rate(labels, confidence_scores):
    sorted_confidence_scores, sorted_labels = zip(*sorted(zip(confidence_scores, labels)))
    sorted_labels = np.array(sorted_labels, dtype=np.int)
    # print('sorted_confidence_scores = ', sorted_confidence_scores)
    # print('sorted_labels = ', sorted_labels)
    all_pass_rates = []
    all_accuracies = []
    for i in range(len(sorted_labels)):
        pass_labels = sorted_labels[i:]
        pass_rate = len(pass_labels) / len(sorted_labels)
        all_pass_rates.append(pass_rate)
        accuracy = np.sum(pass_labels) / len(pass_labels)
        all_accuracies.append(accuracy)

    return all_pass_rates, all_accuracies


def oracle_score(confidence: ConfidenceFeatures):
    label = ConfidenceEstimator.convert_to_labels([confidence])[0]
    # assign confidence scores randomly in (0, 0.5) for incorrect examples and in (0.5, 1) for correct ones.
    # This way, all correct exampels are ranked above all incorrect examples.
    oracle_confidence = label * (np.random.random() / 2 + 0.5) + (1 - label) * (np.random.random() / 2)
    return oracle_confidence


def evaluate_raw(dev_confidences: Iterable[ConfidenceFeatures], featurizer: Callable):
    """
    Evaluates scores directly, instead of feedeing them into a boosted tree
    """
    dev_labels = ConfidenceEstimator.convert_to_labels(dev_confidences)
    dev_avg_logprobs = [featurizer(c) for c in dev_confidences]
    # _max = np.max(dev_avg_logprobs)
    # _min = np.min(dev_avg_logprobs)
    # dev_avg_logprobs = (dev_avg_logprobs - _min) / (_max - _min)
    precision, recall, thresholds = precision_recall_curve(dev_labels, dev_avg_logprobs)
    pass_rate, accuracies = accuracy_at_pass_rate(dev_labels, dev_avg_logprobs)
    return precision, recall, pass_rate, accuracies, thresholds


def parse_argv(parser):
    parser.add_argument(
        '--confidence_path',
        required=True,
        type=str,
        help='The path to the pickle file where the list of ConfidenceFeatures objects is saved',
    )
    parser.add_argument(
        '--eval_metric',
        type=str,
        default='aucpr',
        choices=['aucpr'],
        help='An xgboost metric. The metric which will be used to select the best model on the validation set.',
    )
    parser.add_argument(
        '--dev_split',
        type=float,
        default=0.2,
        help='The portion of the dataset to use for validation. The rest is used to train.',
    )
    parser.add_argument('--seed', type=int, default=123, help='Random seed to use for reproducibility')
    parser.add_argument(
        '--save', required=True, type=str, help='The directory to save the calibrator model and plots after training'
    )
    parser.add_argument(
        '--name_prefix', required=True, type=str, help='A string to prepend to files associated with the calibrator.'
    )
    parser.add_argument(
        '--plot', action='store_true', help='If True, will plot metrics and save them. Requires Matplotlib installation.'
    )
    parser.add_argument(
        '--testing',
        action='store_true',
        help='If True, will change labels so that not all of them are equal. This is only used for testing purposes.',
    )
    parser.add_argument(
        '--fast',
        action='store_true',
        help='If True, will only train calibrators that don\'t use MC dropout. This substantially increases inference speed.',
    )

    # Options to normalize scores for a given threshold
    parser.add_argument(
        '--threshold',
        type=float,
        default=None,
        help='The threshold above (below) which scores are considered to be positive (negative).',
    )
    parser.add_argument('--precision', type=float, default=None, help='Set this if scores should be normalize by precision.')
    parser.add_argument('--recall', type=float, default=None, help='Set this if scores should be normalize by recall.')


class ConfidenceEstimator:
    def __init__(
        self, name: str, featurizers: List[Union[Callable, Tuple[Callable, Callable]]], eval_metric: str, mc_dropout_num: int
    ):
        raise NotImplementedError()

    def convert_to_features(self, confidences: Iterable[ConfidenceFeatures], train: bool = False):
        raise NotImplementedError()

    @staticmethod
    def convert_to_labels(confidences: Iterable[ConfidenceFeatures]):
        labels = []
        for c in confidences:
            labels.append(c[0].label)
        labels = np.array(labels)
        # logger.info('labels = %s', str(labels))
        return labels

    def convert_to_dataset(self, confidences: Iterable[ConfidenceFeatures], train: bool):
        labels = ConfidenceEstimator.convert_to_labels(confidences)
        features = self.convert_to_features(confidences, train)

        return features, labels

    def train_and_validate(self, train_features, train_labels, dev_features, dev_labels):
        raise NotImplementedError()

    def estimate(self, confidences: Iterable[ConfidenceFeatures]):
        raise NotImplementedError()

    def evaluate(self, dev_features, dev_labels):
        raise NotImplementedError()

    def set_normalization_constant(self, c: float):
        logger.info('Setting normalization constant to %.3f', c)
        self.normalization_constant = c

    def normalize_score(self, scores):
        return [s + self.normalization_constant for s in scores]

    def save(self, path: str):
        with open(path, 'wb') as f:
            dill.dump(self, f, protocol=4)

    @staticmethod
    def is_estimator(path: str):
        return path.endswith('.calib')

    @staticmethod
    def load(path: str):
        with open(path, 'rb') as f:
            obj = dill.load(f)
        return obj


class RawConfidenceEstimator(ConfidenceEstimator):
    def __init__(
        self, name: str, featurizers: List[Union[Callable, Tuple[Callable, Callable]]], eval_metric: str, mc_dropout_num: int
    ):
        # don't change the interface, to match the parent class
        assert (
            len(featurizers) == 1
        ), 'RawConfidenceEstimator only works with a single featurizer that outputs a single numerical value'
        self.name = name
        self.featurizer = featurizers[0]
        self.eval_metric = eval_metric
        self.mc_dropout_num = mc_dropout_num

        self.score = 0
        self.normalization_constant = 0

    def estimate(self, confidences: Iterable[ConfidenceFeatures]):
        confidence_scores = self.convert_to_features(confidences)
        confidence_scores = [float(a) for a in confidence_scores]
        confidence_scores = self.normalize_score(confidence_scores)
        return confidence_scores

    def convert_to_features(self, confidences: Iterable[ConfidenceFeatures], train: bool = False):
        features = [self.featurizer(c) for c in confidences]
        return features

    def train_and_validate(self, train_features, train_labels, dev_features, dev_labels):
        # no training to be done
        precision, recall, pass_rate, accuracies, thresholds = self.evaluate(dev_features, dev_labels)
        score = auc(recall, precision)
        self.score = score
        logger.info('best dev set score = %.3f', score)

    def evaluate(self, dev_features, dev_labels):
        confidence_scores = dev_features
        precision, recall, thresholds = precision_recall_curve(dev_labels, confidence_scores)
        pass_rate, accuracies = accuracy_at_pass_rate(dev_labels, confidence_scores)
        return precision, recall, pass_rate, accuracies, thresholds


class TreeConfidenceEstimator(ConfidenceEstimator):
    def __init__(
        self, name: str, featurizers: List[Union[Callable, Tuple[Callable, Callable]]], eval_metric: str, mc_dropout_num: int
    ):
        self.name = name
        self.featurizers = featurizers
        self.eval_metric = eval_metric

        self.model = None
        self.score = 0
        self.normalization_constant = 0
        self.feature_size = 0
        self.normalizer_sub = 0
        self.normalizer_div = 1

        self.mc_dropout_num = mc_dropout_num

    @staticmethod
    def _extract_confidence_scores(model, dev_dataset):
        prediction_probs = model.predict(dev_dataset, ntree_limit=model.best_ntree_limit)
        return prediction_probs

    def _pad_and_normalize(self, features: List, normalize='var', train: bool = False):
        if train or self.feature_size == 0:
            self.feature_size = max([len(f) for f in features])
            logger.info('feature size of the model is set to %d', self.feature_size)
        padded_features = []
        for f in features:
            f = f[: self.feature_size]  # truncate
            padded_features.append(
                np.pad(f, pad_width=(0, self.feature_size - len(f)), constant_values=np.nan, mode='constant')
            )

        padded_features = np.stack(padded_features)
        if train:
            if normalize == 'var':
                mean = np.nanmean(padded_features, axis=0)
                var = np.nanvar(padded_features, axis=0)
                self.normalizer_sub = mean
                self.normalizer_div = np.sqrt(var)
            elif normalize == 'max':
                _max = np.max(padded_features, axis=0)
                _min = np.min(padded_features, axis=0)
                self.normalizer_sub = _min
                self.normalizer_div = _max - _min
            elif normalize == 'none':
                self.normalizer_sub = 0
                self.normalizer_div = 1
            else:
                raise ValueError('Unexpected value for `normalize`')
        padded_features = (padded_features - self.normalizer_sub) / self.normalizer_div
        padded_features[np.isnan(padded_features)] = 0

        return padded_features

    @staticmethod
    def _interleave_features(features_list: List[List]) -> List:
        all_interleaved = []
        for i in range(len(features_list[0])):
            interleaved_length = features_list[0][i].shape[0] * len(features_list)
            interleaved = np.empty((interleaved_length,), dtype=np.float32)
            for j in range(len(features_list)):
                interleaved[j :: len(features_list)] = features_list[j][i]
            all_interleaved.append(interleaved)

        return all_interleaved

    @staticmethod
    def _concatenate(features_list: List[List]) -> List:
        all_concats = []
        for i in range(len(features_list[0])):
            concat = np.concatenate([features_list[j][i].cpu() for j in range(len(features_list))])
            all_concats.append(concat)

        return all_concats

    def convert_to_features(self, confidences: Iterable[ConfidenceFeatures], train: bool = False):
        # TODO check to make sure padding is always on the right hand side, not in the middle of features
        features = []
        for featurizer in self.featurizers:
            if isinstance(featurizer, tuple):
                feature = TreeConfidenceEstimator._interleave_features(
                    [[f(c) for c in confidences] for f in featurizer]
                )  # list of np.arrays
            else:
                feature = [featurizer(c) for c in confidences]  # list of np.arrays
            features.append(feature)
        features = TreeConfidenceEstimator._concatenate(features)
        padded_features = self._pad_and_normalize(features, train=train)
        # print('concatentated features = ', features)
        # print('padded_features = ', padded_features)

        return padded_features

    def _tune_and_train(self, train_dataset, dev_dataset, dev_labels, scale_pos_weight: float):
        # set of all possible hyperparameters
        max_depth = [3, 5, 7, 10, 20, 30, 50]  # the maximum depth of each tree
        eta = [0.02, 0.1, 0.5, 0.7]  # the training step for each iteration
        num_round = [300]

        best_score = 0
        best_model = None
        best_confusion_matrix = None
        best_params = None
        for m, e, n in itertools.product(max_depth, eta, num_round):
            params = {
                'max_depth': m,
                'eta': e,
                'objective': 'binary:logistic',
                'eval_metric': self.eval_metric,
                'scale_pos_weight': scale_pos_weight,
            }
            evals_result = {}
            model = xgb.train(
                params=params,
                dtrain=train_dataset,
                evals=[(dev_dataset, 'dev')],
                num_boost_round=n,
                early_stopping_rounds=50,
                evals_result=evals_result,
                verbose_eval=False,
            )
            # print('evals_result = ', evals_result)
            prediction_probs = TreeConfidenceEstimator._extract_confidence_scores(model, dev_dataset)
            predictions = np.round(np.asarray(prediction_probs))
            accuracy = accuracy_score(dev_labels, predictions)
            score = model.best_score  # evals_result['dev']['aucpr'][-1]#
            logger.info('score=%.3f \t accuracy=%.1f \t best_iteration=%d \t', score, accuracy * 100, model.best_iteration)
            confusion_m = confusion_matrix(dev_labels, predictions)
            if score > best_score:
                best_score = score
                best_model = model
                best_confusion_matrix = confusion_m
                best_params = m, e, n
            best_score = max(best_score, score)

            self.model = best_model
            self.score = best_score

        return best_model, best_score, best_confusion_matrix, best_params

    def estimate(self, confidences: Iterable[ConfidenceFeatures]):
        features, labels = self.convert_to_dataset(confidences, train=False)
        dataset = xgb.DMatrix(data=features, label=labels)
        confidence_scores = TreeConfidenceEstimator._extract_confidence_scores(self.model, dataset)
        confidence_scores = self.normalize_score(confidence_scores)
        return confidence_scores

    def evaluate(self, dev_features, dev_labels):
        dev_dataset = xgb.DMatrix(data=dev_features, label=dev_labels)
        confidence_scores = TreeConfidenceEstimator._extract_confidence_scores(self.model, dev_dataset)
        precision, recall, thresholds = precision_recall_curve(dev_labels, confidence_scores)
        pass_rate, accuracies = accuracy_at_pass_rate(dev_labels, confidence_scores)

        return precision, recall, pass_rate, accuracies, thresholds

    def train_and_validate(self, train_features, train_labels, dev_features, dev_labels):
        train_dataset = xgb.DMatrix(data=train_features, label=train_labels)
        dev_dataset = xgb.DMatrix(data=dev_features, label=dev_labels)
        scale_pos_weight = np.sum(dev_labels) / (np.sum(1 - dev_labels))  # 1s over 0s
        # logger.info('scale_pos_weight = %f', scale_pos_weight)

        best_model, best_score, best_confusion_matrix, best_params = self._tune_and_train(
            train_dataset=train_dataset, dev_dataset=dev_dataset, dev_labels=dev_labels, scale_pos_weight=scale_pos_weight
        )
        logger.info('best dev set score = %.3f', best_score)
        logger.info('best confusion_matrix = %s', str(best_confusion_matrix))
        logger.info('best hyperparameters (max_depth, eta, num_iterations) = %s', str(best_params))


def find_nearest_index(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


slow_feature_sets = [
    # ([mean_drop_seq_prob(0)], 'raw_mean_seq_prob'),
    # ([mean_drop_prob(0)], 'mean_prob'),
    # ([var_drop_prob(0)], 'var_prob'),
    # ([cv_drop_prob(0)], 'cv_drop_prob'),
    # ([probability_that_2_overtakes_1(0)], 'probability_that_2_overtakes_1'),
    # ([diff_mean_drop_probability_2_and_1(0)], 'diff_mean_drop_probability_2_and_1'),
    # ([diff_var_drop_probability_2_and_1(0)], 'diff_var_drop_probability_2_and_1'),
    # ([diff_nodrop_probability_2_and_1(0)], 'diff_nodrop_probability_2_and_1'),
    # ([(mean_drop_logit(0), nodrop_entropies(0))], 'mean + entropy'),
    # ([prediction_length(0), (mean_drop_logit(0), nodrop_entropies(0))], 'prediction_length + mean + entropy'),
    # ([prediction_length(0), (mean_drop_logit(0), nodrop_entropies(0), cv_drop_logit(0))], 'prediction_length + mean + entropy + cv'),
    # ([max_of(nodrop_logit(0)), max_of(nodrop_entropies(0)), max_of(cv_drop_logit(0))], 'max_logit + max_entropy + max_cv'),
    # ([prediction_length(0), max_of(nodrop_logit(0)), max_of(nodrop_entropies(0)), max_of(cv_drop_logit(0))], 'prediction_length + max_logit + max_entropy + max_cv'),
    # ([prediction_length(0), max_of(nodrop_logit(0)), max_of(nodrop_entropies(0)), max_of(cv_drop_logit(0)), max_of(var_drop_logit(0)), min_of(nodrop_logit(0)), min_of(nodrop_entropies(0)), min_of(cv_drop_logit(0)), min_of(var_drop_logit(0))], 'prediction_length + max_logit + max_entropy + max_cv + max_var + min_logit + min_entropy + min_cv + min_var'),
    # ([prediction_length(0), max_of(nodrop_logit(0)), max_of(nodrop_entropies(0)), max_of(cv_drop_logit(0)), min_of(nodrop_logit(0)), min_of(nodrop_entropies(0)), min_of(cv_drop_logit(0))], 'prediction_length + max_logit + max_entropy + max_cv + min_logit + min_entropy + min_cv'),
    # ([nodrop_avg_logprob(0), prediction_length(0), max_of(nodrop_logit(0)), max_of(nodrop_entropies(0)), max_of(cv_drop_logit(0)), min_of(nodrop_logit(0)), input_length(0)], 'logprob + prediction_length + max_logit + max_entropy + max_cv + min_logit + input_length'),
    # ([nodrop_seq_prob(0), prediction_length(0), max_of(mean_drop_prob(0)), max_of(nodrop_entropies(0)), max_of(cev_drop_prob(0)), min_of(mean_drop_prob(0)), input_length(0), cev_drop_seq_prob(0), mean_drop_seq_prob(0)], 'prob + prediction_length + max_logit + max_entropy + max_cv + min_logit + input_length + cev_seq_prob + mean_seq_prob'),
    # ([variance_of_beam_logits], 'var_beam_logits'),
    # ([variance_of_beam_probs], 'var_beam_probs'),
    # ([mean_drop_avg_logprob(0)], 'mean_drop_avg_logprob'),
    # ([var_drop_avg_logprob(0)], 'var_drop_avg_logprob'),
    # ([cv_drop_avg_logprob(0)], 'cv_drop_avg_logprob'),
    # One of these three usually outperforms all the other ones:
    (
        [
            prediction_length(0),
            max_of(nodrop_logit(0)),
            max_of(nodrop_entropies(0)),
            max_of(cv_drop_logit(0)),
            min_of(nodrop_logit(0)),
            min_of(nodrop_entropies(0)),
            min_of(cv_drop_logit(0)),
            input_length(0),
        ],
        'prediction_length + max_logit + max_entropy + max_cv + min_logit + min_entropy + min_cv + input_length',
    ),
    (
        [
            nodrop_avg_logprob(0),
            prediction_length(0),
            max_of(nodrop_logit(0)),
            max_of(nodrop_entropies(0)),
            max_of(cv_drop_logit(0)),
            min_of(nodrop_logit(0)),
            min_of(nodrop_entropies(0)),
            min_of(cv_drop_logit(0)),
            input_length(0),
        ],
        'logprob + prediction_length + max_logit + max_entropy + max_cv + min_logit + min_entropy + min_cv + input_length',
    ),
    (
        [
            nodrop_avg_logprob(0),
            prediction_length(0),
            max_of(nodrop_logit(0)),
            max_of(nodrop_entropies(0)),
            max_of(cv_drop_logit(0)),
            min_of(nodrop_logit(0)),
            input_length(0),
            cev_drop_seq_prob(0),
            mean_drop_seq_prob(0),
        ],
        'logprob + prediction_length + max_logit + max_entropy + max_cv + min_logit + input_length + cev_seq_prob + mean_seq_prob',
    ),
]

fast_feature_sets = [
    # ([oracle_score], 'raw_oracle'),
    # ([nodrop_avg_logprob(0)], 'raw_avg_logprob'),
    ([nodrop_seq_prob(0)], 'raw_seq_prob'),
    # ([neg_of(var_drop_seq_prob(0))], 'raw_var_seq_prob'),
    # ([neg_of(cv_drop_seq_prob(0))], 'raw_cv_seq_prob'),
    # ([cev_drop_seq_prob(0)], 'raw_cev_drop_seq_prob'),
    # ([nodrop_entropies(0)], 'entropy'),
    # These three are the fast versions of the best slow feature sets
    (
        [
            prediction_length(0),
            max_of(nodrop_logit(0)),
            max_of(nodrop_entropies(0)),
            min_of(nodrop_logit(0)),
            min_of(nodrop_entropies(0)),
            input_length(0),
        ],
        'prediction_length + max_logit + max_entropy  + min_logit + min_entropy + input_length',
    ),
    (
        [
            nodrop_avg_logprob(0),
            prediction_length(0),
            max_of(nodrop_logit(0)),
            max_of(nodrop_entropies(0)),
            min_of(nodrop_logit(0)),
            min_of(nodrop_entropies(0)),
            input_length(0),
        ],
        'logprob + prediction_length + max_logit + max_entropy + min_logit + min_entropy + input_length',
    ),
    (
        [
            nodrop_avg_logprob(0),
            prediction_length(0),
            max_of(nodrop_logit(0)),
            max_of(nodrop_entropies(0)),
            min_of(nodrop_logit(0)),
            input_length(0),
        ],
        'logprob + prediction_length + max_logit + max_entropy + min_logit + input_length',
    ),
]


def main(args):

    if args.threshold is not None:
        assert (args.precision is not None and args.recall is None) or (
            args.precision is None and args.recall is not None
        ), 'When `--threshold` is specified, exactly one of `--precision` and `--recall` should be set.'

    if args.plot:
        from matplotlib import pyplot  # lazy import

    confidences = torch.load(args.confidence_path, map_location=torch.device('cuda' if torch.cuda.is_available() else 'cpu'))

    all_estimators = []
    train_confidences, dev_confidences = train_test_split(confidences, test_size=args.dev_split, random_state=args.seed)

    feature_sets = fast_feature_sets
    if not args.fast:
        feature_sets += slow_feature_sets
    for f, name in feature_sets:
        if name.startswith('raw'):
            estimator_class = RawConfidenceEstimator
        else:
            estimator_class = TreeConfidenceEstimator
        mc_dropout_num = train_confidences[0][0].mc_dropout_num
        estimator = estimator_class(name=name, featurizers=f, eval_metric=args.eval_metric, mc_dropout_num=mc_dropout_num)
        logger.info('name = %s', name)

        train_features, train_labels = estimator.convert_to_dataset(train_confidences, train=True)
        dev_features, dev_labels = estimator.convert_to_dataset(dev_confidences, train=False)
        if args.testing:
            if train_labels.all() or (~train_labels).all():
                train_labels[0] = ~train_labels[0]
            if dev_labels.all() or (~dev_labels).all():
                dev_labels[0] = ~dev_labels[0]
        estimator.train_and_validate(train_features, train_labels, dev_features, dev_labels)
        precision, recall, pass_rate, accuracies, thresholds = estimator.evaluate(dev_features, dev_labels)
        if args.threshold:
            # set the threshold using dev set
            if args.recall:
                threshold = thresholds[find_nearest_index(recall, args.recall)]
            else:
                assert args.precision
                threshold = thresholds[find_nearest_index(precision, args.precision)]
            estimator.set_normalization_constant(args.threshold - threshold)

        if args.plot:
            pyplot.figure('precision-recall')
            pyplot.plot(recall, precision, marker='.', label=name)
            pyplot.figure('thresholds')
            pyplot.plot(range(len(thresholds)), thresholds, marker='*', label=name + ' (thresholds)')
            pyplot.figure('pass_rate')
            pyplot.plot(pass_rate, accuracies, marker='.', label=name)

        all_estimators.append(estimator)

    logger.info('\n' + '\n'.join([f'{e.name}: {e.score:.3f}' for e in all_estimators]))
    best_estimator = all_estimators[np.argmax([e.score for e in all_estimators])]
    logger.info('Best estimator is %s with score = %.3f', best_estimator.name, best_estimator.score)
    best_estimator.save(os.path.join(args.save, args.name_prefix + '.calib'))

    if args.plot:
        pyplot.figure('precision-recall')
        pyplot.legend(prop={'size': 6})
        pyplot.grid()
        pyplot.xticks(np.arange(0, 1, 0.1))
        pyplot.xlim(0, 1)
        pyplot.xlabel('Recall')
        pyplot.ylabel('Precision')
        pyplot.savefig(os.path.join(args.save, args.name_prefix + '_precision-recall.svg'))

        pyplot.figure('thresholds')
        pyplot.legend(prop={'size': 6})
        pyplot.grid()
        pyplot.xlabel('Index')
        pyplot.ylabel('Confidence Threshold')
        pyplot.savefig(os.path.join(args.save, args.name_prefix + '_threshold.svg'))

        pyplot.figure('pass_rate')
        pyplot.legend(prop={'size': 6})
        pyplot.grid()
        pyplot.xticks(np.arange(0, 1, 0.1))
        pyplot.xlim(0, 1)
        pyplot.xlabel('Pass Rate')
        pyplot.ylabel('Accuracy')
        pyplot.savefig(os.path.join(args.save, args.name_prefix + '_pass-accuracy.svg'))



import glob
import logging
import math
import os
import re
import shutil

import torch
from transformers.models.mbart.tokenization_mbart50 import FAIRSEQ_LANGUAGE_CODES

from ..metrics import computeBLEU
from ..util import get_mbart_lang
from .transformers_utils import MARIAN_GROUP_MEMBERS

logger = logging.getLogger(__name__)


def shift_tokens_right(input_ids, pad_token_id):
    """
    Shift input ids one token to the right, and wrap the last non pad token (usually <eos>).
    Adopted from huggingface's finetune.py code
    """
    prev_output_tokens = input_ids.clone()
    index_of_eos = (input_ids.ne(pad_token_id).sum(dim=1) - 1).unsqueeze(-1)
    prev_output_tokens[:, 0] = input_ids.gather(1, index_of_eos).squeeze()
    prev_output_tokens[:, 1:] = input_ids[:, :-1]
    return prev_output_tokens


def freeze_params(model):
    for par in model.parameters():
        par.requires_grad = False


def unfreeze_params(model):
    for par in model.parameters():
        par.requires_grad = True


def freeze_embeds(model):
    """Freeze token embeddings and positional embeddings for bart, just token embeddings for t5."""
    if hasattr(model, 'model'):
        freeze_params(model.model.shared)
        for d in [model.model.encoder, model.model.decoder]:
            freeze_params(d.embed_positions)
            freeze_params(d.embed_tokens)
    else:
        freeze_params(model.shared)
        for d in [model.encoder, model.decoder]:
            freeze_params(d.embed_tokens)


def unfreeze_embeds(model):
    """Unfreeze token embeddings and positional embeddings for bart, just token embeddings for t5."""
    if hasattr(model, 'model'):
        unfreeze_params(model.model.shared)
        for d in [model.model.encoder, model.model.decoder]:
            unfreeze_params(d.embed_positions)
            unfreeze_params(d.embed_tokens)
    else:
        unfreeze_params(model.shared)
        for d in [model.encoder, model.decoder]:
            unfreeze_params(d.embed_tokens)


def check_args(args):
    if args.model_type == 'marian' and args.model_name_or_path.rsplit('-', 1)[1] in MARIAN_GROUP_MEMBERS:
        if not args.tgt_lang:
            raise ValueError(
                'For translation task using Marian model, if target language is a group of languages, '
                'you have to specify the --tgt_lang flag.'
            )
        elif args.tgt_lang not in MARIAN_GROUP_MEMBERS[args.model_name_or_path.rsplit('-', 1)[1]]:
            if args.tgt_lang == 'pl':
                args.tgt_lang = 'pol'
            elif args.tgt_lang == 'fa':
                args.tgt_lang = 'pes'
            else:
                raise ValueError(
                    'Target language is not in the model group languages, please specify the correct target language.'
                )

    if args.model_type == 'marian' and args.model_name_or_path.rsplit('-', 2)[1] in MARIAN_GROUP_MEMBERS:
        if not args.src_lang:
            raise ValueError(
                'For translation task using Marian model, if source language is a group of languages, '
                'you have to specify the --src_lang flag.'
            )
        elif args.src_lang not in MARIAN_GROUP_MEMBERS[args.model_name_or_path.rsplit('-', 2)[1]]:
            if args.src_lang == 'pl':
                args.src_lang = 'pol'
            elif args.src_lang == 'fa':
                args.src_lang = 'pes'
            raise ValueError(
                'Source language is not in the model group languages, please specify the correct source language.'
            )

    if args.model_type == 'marian' and args.model_name_or_path.rsplit('-', 1)[1] not in MARIAN_GROUP_MEMBERS and args.tgt_lang:
        logger.warning(
            'Target language should not be provided when using Marian models with single target language,'
            ' otherwise the translation outputs will be incorrect; thus we ignore the target language you provided...'
        )
        args.tgt_lang = None

    if args.model_type == 'marian' and args.model_name_or_path.rsplit('-', 2)[1] not in MARIAN_GROUP_MEMBERS and args.src_lang:
        logger.warning(
            'Source language should not be provided when using Marian models with single source language,'
            ' otherwise the translation outputs will be incorrect; thus we ignore the source language you provided...'
        )
        args.src_lang = None

    if args.model_type == 'mbart' and not (args.tgt_lang and args.src_lang):
        raise ValueError('Source and Target language should be provided when using mBART cc25 model')

    # adjust language ids for mbart models
    if args.model_type in ['mbart', 'mbart50']:
        if args.src_lang not in FAIRSEQ_LANGUAGE_CODES:
            args.src_lang = get_mbart_lang(args.src_lang)
        if args.tgt_lang not in FAIRSEQ_LANGUAGE_CODES:
            args.tgt_lang = get_mbart_lang(args.tgt_lang)


def sort_checkpoints(output_dir):
    return list(sorted(glob.glob(os.path.join(output_dir, "checkpointepoch=*.ckpt"), recursive=True)))


def get_transformer_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, dimension):
    num_warmup_steps = max(1, num_warmup_steps)

    def lr_lambda(current_step):
        current_step += 1
        return (
            1.0
            / math.sqrt(dimension)
            * min(1 / math.sqrt(current_step), current_step / (num_warmup_steps * math.sqrt(num_warmup_steps)))
        )

    return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)


def _rotate_checkpoints(args, checkpoint_prefix, use_mtime=False):
    if not args.save_total_limit:
        return
    if args.save_total_limit <= 0:
        return

    # Check if we should delete older checkpoint(s)
    glob_checkpoints = glob.glob(os.path.join(args.output_dir, '{}-*'.format(checkpoint_prefix)))
    if len(glob_checkpoints) <= args.save_total_limit:
        return

    ordering_and_checkpoint_path = []
    for path in glob_checkpoints:
        if use_mtime:
            ordering_and_checkpoint_path.append((os.path.getmtime(path), path))
        else:
            regex_match = re.match('.*{}-([0-9]+)'.format(checkpoint_prefix), path)
            if regex_match and regex_match.groups():
                ordering_and_checkpoint_path.append((int(regex_match.groups()[0]), path))

    checkpoints_sorted = sorted(ordering_and_checkpoint_path)
    checkpoints_sorted = [checkpoint[1] for checkpoint in checkpoints_sorted]
    number_of_checkpoints_to_delete = max(0, len(checkpoints_sorted) - args.save_total_limit)
    checkpoints_to_be_deleted = checkpoints_sorted[:number_of_checkpoints_to_delete]
    for checkpoint in checkpoints_to_be_deleted:
        logger.info("Deleting older checkpoint [{}] due to args.save_total_limit".format(checkpoint))
        shutil.rmtree(checkpoint)


def compute_metrics(generations, golds, reduction='average'):
    """
    Inputs:
        generations: a list of list of strings; generations[i] is a list of all generated outputs of the model for example i
        golds: a list of strings; golds[i] is the gold answer for example i
        reduction: how we should compute an example's metrics from its multiple generations
    """
    total_bleu = 0.0
    # all_bleu = []
    total_exact_match = 0.0
    count = 0.0
    for idx, output in enumerate(generations):
        bleu_score = 0.0
        exact_match = 0.0
        for sample in output:
            if reduction == 'average':
                bleu_score += computeBLEU([sample], [[golds[idx]]])
            else:
                bleu_score = max(bleu_score, computeBLEU([sample], [[golds[idx]]]))
            if re.sub('\s+', '', sample).lower() == re.sub('\s+', '', golds[idx]).lower():
                if reduction == 'average':
                    exact_match += 1
                else:
                    exact_match = max(exact_match, 1)
        if reduction == 'average':
            bleu_score /= len(output)
            exact_match /= len(output)
        total_bleu += bleu_score
        total_exact_match += exact_match
        count += 1

    return {'bleu': total_bleu / count, 'em': total_exact_match / count * 100}



#
# Copyright (c) 2020-2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
import os
import shutil
import time
from pprint import pformat

import ujson

from .arguments import post_parse_general, save_args
from .data_utils.bootleg import Bootleg
from .util import set_seed

logger = logging.getLogger(__name__)


def parse_argv(parser):
    parser.add_argument('--root', default='.', type=str, help='root directory for data, results, embeddings, code, etc.')
    parser.add_argument('--save', required=True, type=str, help='where to save results.')
    parser.add_argument('--embeddings', default='.embeddings/', type=str, help='where to save embeddings.')
    parser.add_argument('--data', default='.data/', type=str, help='where to load data from.')
    parser.add_argument('--cache', default='.cache/', type=str, help='where to save cached files')

    parser.add_argument(
        '--train_languages',
        type=str,
        default='en',
        dest='train_src_languages',
        help='Specify dataset source languages used during training for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )
    parser.add_argument(
        '--eval_languages',
        type=str,
        default='en',
        dest='eval_src_languages',
        help='Specify dataset source languages used during validation for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )

    parser.add_argument(
        '--train_tgt_languages',
        type=str,
        default='en',
        help='Specify dataset target languages used during training for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )
    parser.add_argument(
        '--eval_tgt_languages',
        type=str,
        default='en',
        help='Specify dataset target languages used during validation for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )

    parser.add_argument(
        '--train_tasks', nargs='+', type=str, dest='train_task_names', help='tasks to use for training', required=True
    )

    parser.add_argument(
        '--val_tasks', nargs='+', type=str, dest='val_task_names', help='tasks to collect evaluation metrics for'
    )

    parser.add_argument(
        '--val_batch_size',
        nargs='+',
        default=[256],
        type=int,
        help='Batch size for validation corresponding to tasks in val tasks',
    )

    parser.add_argument(
        '--sentence_batching', action='store_true', help='Batch same sentences together (used for multilingual tasks)'
    )
    parser.add_argument(
        '--train_batch_size',
        type=int,
        default=0,
        help='Number of samples to use in each batch; will be used instead of train_batch_tokens when sentence_batching is on',
    )

    parser.add_argument('--eval_set_name', type=str, help='Evaluation dataset name to use during training')

    parser.add_argument('--subsample', default=20000000, type=int, help='subsample the datasets')
    parser.add_argument('--preserve_case', action='store_false', dest='lower', help='whether to preserve casing for all text')

    parser.add_argument(
        "--almond_has_single_program",
        action='store_false',
        dest='almond_has_multiple_programs',
        help='Indicate if almond dataset has multiple programs for each sentence',
    )

    parser.add_argument(
        '--num_workers', type=int, default=0, help='Number of processes to use for data loading (0 means no multiprocessing)'
    )

    parser.add_argument(
        '--database_type', default='json', choices=['json', 'remote-elastic'], help='database to interact with for NER'
    )

    parser.add_argument(
        '--min_entity_len', type=int, default=1, help='Minimum length for entities when ngrams database_lookup_method is used '
    )
    parser.add_argument(
        '--max_entity_len', type=int, default=4, help='Maximum length for entities when ngrams database_lookup_method is used '
    )
    parser.add_argument(
        '--database_dir',
        type=str,
        default='database/',
        help='Database folder containing all relevant files (e.g. alias2qids, pretrained models for bootleg)',
    )

    parser.add_argument(
        '--bootleg_output_dir',
        type=str,
        default='results_temp',
        help='Path to folder where bootleg prepped files should be saved',
    )
    parser.add_argument('--bootleg_model', type=str, default='bootleg_uncased_mini', help='Bootleg model to use')
    parser.add_argument(
        '--bootleg_dump_mode',
        choices=['dump_preds', 'dump_embs'],
        default='dump_preds',
        help='dump_preds will dump only predictions; dump_embs will dump both prediction and embeddings',
    )
    parser.add_argument('--bootleg_batch_size', type=int, default=32, help='Batch size used for inference using bootleg')
    parser.add_argument(
        '--bootleg_prob_threshold',
        type=float,
        default=0.3,
        help='Probability threshold for accepting a candidate for a mention',
    )
    parser.add_argument(
        '--bootleg_dataset_threads',
        type=int,
        default=1,
        help='Number of threads for parallel processing of dataset in bootleg',
    )
    parser.add_argument(
        '--bootleg_dataloader_threads',
        type=int,
        default=1,
        help='Number of threads for parallel loading of datasets in bootleg',
    )
    parser.add_argument(
        '--bootleg_extract_num_workers', type=int, default=32, help='Number of workers for extracing mentions step of bootleg'
    )
    parser.add_argument('--bootleg_post_process_types', action='store_true', help='Postprocess bootleg types')

    parser.add_argument('--verbose', action='store_true', help='Print detected types for each token')
    parser.add_argument(
        '--almond_domains', nargs='+', default=[], help='Domains used for almond dataset; e.g. music, books, ...'
    )
    parser.add_argument(
        '--ned_features',
        nargs='+',
        type=str,
        default=['type_id', 'type_prob'],
        help='Features that will be extracted for each entity: "type" and "freq" are supported.' ' Order is important',
    )
    parser.add_argument(
        '--ned_features_size',
        nargs='+',
        type=int,
        default=[1, 1],
        help='Max length of each feature vector. All features are padded up to this length',
    )
    parser.add_argument(
        '--ned_features_default_val',
        nargs='+',
        type=float,
        default=[0, 1.0],
        help='Max length of each feature vector. All features are padded up to this length',
    )

    parser.add_argument(
        '--almond_lang_as_question',
        action='store_true',
        help='if true will use "Translate from ${language} to ThingTalk" for question',
    )
    parser.add_argument(
        '--almond_detokenize_sentence',
        action='store_true',
        help='undo word tokenization of almond sentence fields (useful if the tokenizer is sentencepiece)',
    )
    parser.add_argument(
        '--almond_thingtalk_version', type=int, choices=[1, 2], default=2, help='Thingtalk version for almond datasets'
    )

    parser.add_argument('--seed', default=123, type=int, help='Random seed.')
    parser.add_argument('--devices', default=[0], nargs='+', type=int, help='a list of devices that can be used for training')

    parser.add_argument(
        '--no_commit',
        action='store_false',
        dest='commit',
        help='do not track the git commit associated with this training run',
    )
    parser.add_argument(
        '--exist_ok', action='store_true', help='Ok if the save directory already exists, i.e. overwrite is ok'
    )

    parser.add_argument(
        '--skip_cache', action='store_true', help='whether to use exisiting cached splits or generate new ones'
    )
    parser.add_argument(
        '--cache_input_data', action='store_true', help='Cache examples from input data for faster subsequent trainings'
    )

    parser.add_argument('--use_curriculum', action='store_true', help='Use curriculum learning')
    parser.add_argument(
        '--aux_dataset', default='', type=str, help='path to auxiliary dataset (ignored if curriculum is not used)'
    )
    parser.add_argument("--add_types_to_text", default='no', choices=['no', 'insert', 'append'])

    # token classification task args
    parser.add_argument('--num_labels', type=int, help='num_labels for classification tasks')
    parser.add_argument('--ner_domains', nargs='+', type=str, help='domains to use for CrossNER task')
    parser.add_argument(
        '--hf_test_overfit',
        action='store_true',
        help='Debugging flag for hf datasets where validation will be performed on train set',
    )


def bootleg_process_splits(args, examples, path, task, bootleg, mode='train'):
    config_overrides = bootleg.fixed_overrides

    input_file_dir = os.path.dirname(path)
    input_file_name = os.path.basename(path.rsplit('.', 1)[0] + '_bootleg.jsonl')

    data_overrides = ["--data_config.data_dir", input_file_dir, "--data_config.test_dataset.file", input_file_name]

    # get config args
    config_overrides.extend(data_overrides)
    config_args = bootleg.create_config(config_overrides)

    if mode == 'dump':
        # create jsonl files from input examples
        # jsonl is the input format bootleg expects
        bootleg.create_jsonl(path, examples, task.utterance_field())

        # extract mentions and mention spans in the sentence and write them to output jsonl files
        bootleg.extract_mentions(path)

        # find the right entity candidate for each mention
        bootleg.disambiguate_mentions(config_args)

    # extract features for each token in input sentence from bootleg outputs
    all_token_type_ids, all_tokens_type_probs = bootleg.collect_features(
        input_file_name[: -len('_bootleg.jsonl')], args.subsample, getattr(task, 'TTtype2qid', None)
    )

    all_token_type_ids = all_token_type_ids[: args.subsample]
    all_tokens_type_probs = all_tokens_type_probs[: args.subsample]

    # override examples features with bootleg features
    if mode != 'dump':
        assert len(examples) == len(all_token_type_ids) == len(all_tokens_type_probs)
        for n, (ex, tokens_type_ids, tokens_type_probs) in enumerate(zip(examples, all_token_type_ids, all_tokens_type_probs)):
            if task.utterance_field() == 'question':
                for i in range(len(tokens_type_ids)):
                    examples[n].question_feature[i].type_id = tokens_type_ids[i]
                    examples[n].question_feature[i].type_prob = tokens_type_probs[i]
                question_plus_types = task.add_type_tokens(ex.question, ex.question_feature, args.add_types_to_text)
                examples[n].question_plus_types = question_plus_types

            else:
                # context is the utterance field
                for i in range(len(tokens_type_ids)):
                    examples[n].context_feature[i].type_id = tokens_type_ids[i]
                    examples[n].context_feature[i].type_prob = tokens_type_probs[i]
                context_plus_types = task.add_type_tokens(ex.context, ex.context_feature, args.add_types_to_text)
                examples[n].context_plus_types = context_plus_types

    if args.verbose:
        for ex in examples:
            print()
            print(
                *[
                    f'context token: {token}\ttype: {token_type}'
                    for token, token_type in zip(ex.context.split(' '), ex.context_plus_feature)
                ],
                sep='\n',
            )
            print(
                *[
                    f'question token: {token}\ttype: {token_type}'
                    for token, token_type in zip(ex.question.split(' '), ex.question_plus_feature)
                ],
                sep='\n',
            )


def dump_bootleg_features(args, logger):

    bootleg = Bootleg(args)

    train_eval_shared_kwargs = {
        'subsample': args.subsample,
        'skip_cache': args.skip_cache,
        'cache_input_data': args.cache_input_data,
        'sentence_batching': args.sentence_batching,
        'almond_lang_as_question': args.almond_lang_as_question,
        'num_workers': args.num_workers,
    }

    assert len(args.train_tasks) == len(args.val_tasks)

    for train_task, val_task in zip(args.train_tasks, args.val_tasks):

        # process train split
        logger.info(f'Loading {train_task.name}')
        kwargs = {'test': None, 'validation': None}
        kwargs.update(train_eval_shared_kwargs)
        kwargs['all_dirs'] = args.train_src_languages
        kwargs['cached_path'] = os.path.join(args.cache, train_task.name)
        kwargs['ner_domains'] = args.ner_domains
        if args.use_curriculum:
            kwargs['curriculum'] = True

        logger.info(f'Adding {train_task.name} to training datasets')
        t0 = time.time()
        splits, paths = train_task.get_splits(args.data, lower=args.lower, **kwargs)
        t1 = time.time()
        logger.info('Data loading took {:.2f} seconds'.format(t1 - t0))
        assert not splits.eval and not splits.test
        assert not args.use_curriculum

        train_dataset = splits.train
        train_path = paths.train

        logger.info(f'{train_task.name} has {len(splits.train)} training examples')

        if hasattr(train_task, 'all_schema_types'):
            logger.info(f'train all_schema_types: {train_task.all_schema_types}')

        if train_task.name.startswith('almond'):
            if args.ned_features_default_val:
                args.db_unk_id = int(args.ned_features_default_val[0])
            else:
                args.db_unk_id = 0
            if args.do_ned:
                if bootleg:
                    args.num_db_types = len(bootleg.typeqid2id)
                elif getattr(train_task, 'db', None):
                    args.num_db_types = len(train_task.db.typeqid2id)
            else:
                args.num_db_types = 0
        else:
            args.db_unk_id = 0
            args.num_db_types = 0
        save_args(args, force_overwrite=True)

        # process validation split
        logger.info(f'Loading {val_task.name}')
        kwargs = {'train': None, 'test': None}
        # choose best model based on this dev set
        if args.eval_set_name is not None:
            kwargs['validation'] = args.eval_set_name
        kwargs.update(train_eval_shared_kwargs)
        kwargs['all_dirs'] = args.eval_src_languages
        kwargs['cached_path'] = os.path.join(args.cache, val_task.name)
        kwargs['ner_domains'] = args.ner_domains
        kwargs['hf_test_overfit'] = args.hf_test_overfit

        logger.info(f'Adding {val_task.name} to validation datasets')
        splits, paths = val_task.get_splits(args.data, lower=args.lower, **kwargs)

        assert not splits.train and not splits.test and not splits.aux
        logger.info(f'{val_task.name} has {len(splits.eval)} validation examples')

        if hasattr(val_task, 'all_schema_types'):
            logger.info(f'eval all_schema_types: {val_task.all_schema_types}')

        eval_dataset = splits.eval

        # merge all splits before feeding to bootleg
        all_examples = train_dataset.examples + eval_dataset.examples
        dir_name = os.path.dirname(train_path)
        extension = train_path.rsplit('.', 1)[1]
        all_paths = os.path.join(dir_name, 'combined' + '.' + extension)

        assert train_task == val_task
        bootleg_process_splits(args, all_examples, all_paths, train_task, bootleg, mode='dump')

        eval_file_name = args.eval_set_name if args.eval_set_name is not None else 'eval'

        train_output_path = f'{args.bootleg_output_dir}/train_bootleg/{bootleg.ckpt_name}'
        eval_output_path = f'{args.bootleg_output_dir}/{eval_file_name}_bootleg/{bootleg.ckpt_name}'
        os.makedirs(train_output_path, exist_ok=True)
        os.makedirs(eval_output_path, exist_ok=True)
        train_output_file = open(os.path.join(train_output_path, 'bootleg_labels.jsonl'), 'w')
        eval_output_file = open(os.path.join(eval_output_path, 'bootleg_labels.jsonl'), 'w')

        train_size = len(train_dataset.examples)
        eval_size = len(eval_dataset.examples)

        # unmerge bootleg dumped labels
        with open(f'{args.bootleg_output_dir}/combined_bootleg/{bootleg.ckpt_name}/bootleg_labels.jsonl', 'r') as fin:

            # sort output lines first to align with input (required for bootleg >=1.0.0)
            all_lines = fin.readlines()
            all_sent_ids = [ujson.loads(line)['sent_idx_unq'] for line in all_lines]
            all_lines = list(zip(*sorted(zip(all_sent_ids, all_lines), key=lambda item: item[0])))[1]

            i = 0
            for line in all_lines:
                if i < train_size:
                    train_output_file.write(line)
                else:
                    eval_output_file.write(line)
                i += 1

        assert i == train_size + eval_size

        # close output files
        train_output_file.close()
        eval_output_file.close()

        shutil.rmtree(f'{args.bootleg_output_dir}/combined_bootleg')

    logger.info('Created bootleg features for provided datasets with subsampling: {}'.format(args.subsample))


def main(args):

    args.do_ned = True
    args.ned_retrieve_method = 'bootleg'
    args.override_context = None
    args.override_question = None

    args = post_parse_general(args)
    set_seed(args)

    logger.info(f'Arguments:\n{pformat(vars(args))}')

    dump_bootleg_features(args, logger)
